from fastapi.testclient import TestClient


def test_simple_department(client: TestClient):
    response = client.get(
        "/api/v1/mssql/mssql_department?format=json&limit=50",
    )
    assert response.status_code == 200
    departments = response.json()
    assert len(departments) == 16


def test_filter_group_name(client: TestClient):
    response = client.get(
        "/api/v1/mssql/mssql_department?format=json&limit=100&GroupName=Research%20and%20Development",
    )
    assert response.status_code == 200
    tables = response.json()
    assert len(tables) == 3


def test_filter_offset(client: TestClient):
    response = client.get(
        "/api/v1/mssql/mssql_department?format=json&limit=100&offset=10",
    )
    assert response.status_code == 200
    tables = response.json()
    assert len(tables) == 6


def test_metadata_detail(client: TestClient):
    response = client.get(
        "/api/v1/mssql/mssql_department/metadata_detail",
    )
    assert response.status_code == 200



from typing import Callable, Literal, TYPE_CHECKING
import fsspec
import adlfs
import os
import urllib.parse

if TYPE_CHECKING:
    from azure.core.credentials import TokenCredential


def _convert_options(
    uri: str,
    options: dict | None,
    flavor: Literal["fsspec", "object_store", "original"],
    token_retrieval_func: "Callable[[str], TokenCredential] | None" = None,
):
    if options is None:
        return uri, None
    if flavor == "fsspec":
        from deltalake2db.azure_helper import get_storage_options_fsspec

        return uri, get_storage_options_fsspec(options)
    elif flavor == "original":
        return uri, options
    else:
        from deltalake2db.azure_helper import get_storage_options_object_store

        nr, no = get_storage_options_object_store(uri, options, token_retrieval_func)
        assert isinstance(nr, str)
        return nr, no


local_versions = dict()


class SourceUri:
    uri: str
    account: str | None

    def __init__(
        self,
        uri: str,
        account: str | None,
        accounts: dict,
        data_path: str | None,
        token_retrieval_func: "Callable[[SourceUri, str], TokenCredential] | None" = None,
    ):
        self.uri = uri
        self.account = account
        self.accounts = accounts or {}
        self.data_path = data_path
        self.token_retrieval_func = token_retrieval_func
        self.retrieve_token = (
            (lambda v: token_retrieval_func(self, v)) if token_retrieval_func else None
        )
        self.real_uri = (
            uri
            if "://" in uri or account is not None or data_path is None
            else os.path.join(data_path, uri)
        )

    def is_azure(self):
        return (
            self.uri.startswith("azure://")  # duckdb
            or self.uri.startswith("az://")  # duckdb
            or self.uri.startswith("abfs://")  # fsspec
            or self.uri.startswith("abfss://")  # fsspec
        )

    def get_fs_spec(self) -> tuple[fsspec.AbstractFileSystem, str]:
        if self.account is None:
            return fsspec.filesystem("file"), self.real_uri
        real_uri = self.real_uri
        real_uri, opts = _convert_options(
            real_uri,
            self.accounts.get(self.account, {}),
            "fsspec",
            token_retrieval_func=self.retrieve_token,
        )
        assert opts is not None
        if self.is_azure():
            return adlfs.AzureBlobFileSystem(**opts), real_uri  # type: ignore
        else:
            pr = urllib.parse.urlparse(self.uri)
            return fsspec.filesystem(pr, **opts), real_uri

    def get_uri_options(
        self, *, flavor: Literal["fsspec", "object_store", "original"]
    ) -> tuple[str, dict | None]:
        return _convert_options(
            self.real_uri,
            self.accounts.get(self.account) if self.account else None,
            flavor,
            token_retrieval_func=self.retrieve_token,
        )

    def exists(self) -> bool:
        if self.account is None:
            return os.path.exists(self.real_uri)
        fs, fs_path = self.get_fs_spec()
        return fs.exists(fs_path)

    def copy_to_local(self, local_path: str, delta_table: bool):
        if self.account is None:
            raise ValueError("Cannot copy local files")
        if not delta_table:
            fs, fs_path = self.get_fs_spec()
            stat = fs.stat(fs_path)
            lm = stat.get("last_modified", stat.get("Last-Modified"))
            if not lm:
                raise ValueError("Cannot determine last modified date", stat.keys())

            if local_versions.get(self.uri) != lm:
                fs.get_file(fs_path, local_path)
                local_versions[self.uri] = lm
            return SourceUri(
                uri=local_path,
                data_path=None,
                account=None,
                accounts=self.accounts,
                token_retrieval_func=self.token_retrieval_func,
            )
        if delta_table:
            from deltalake2db.delta_meta_retrieval import get_meta, PolarsEngine

            df_uri, df_opts = self.get_uri_options(flavor="object_store")
            meta = get_meta(PolarsEngine(df_opts), df_uri)
            vnr = meta.version
            if local_versions.get(self.uri) == vnr:
                return SourceUri(
                    uri=local_path,
                    data_path=None,
                    account=None,
                    accounts=self.accounts,
                    token_retrieval_func=self.token_retrieval_func,
                )

            os.makedirs(local_path + "/_delta_log", exist_ok=True)
            fs, fs_path = self.get_fs_spec()
            if fs.exists(fs_path.removesuffix("/") + "/_delta_log"):
                fs.get(
                    fs_path + "/_delta_log/",
                    local_path + "/_delta_log/",
                    recursive=True,
                )
                for path in meta.add_actions.keys():
                    if not os.path.exists(local_path + "/" + path):
                        fs.get_file(fs_path + "/" + path, local_path + "/" + path)
            else:
                fs.get(fs_path, local_path, recursive=True)
            local_versions[self.uri] = vnr
            return SourceUri(
                uri=local_path,
                data_path=None,
                account=None,
                accounts=self.accounts,
                token_retrieval_func=self.token_retrieval_func,
            )

    def __str__(self):
        return self.real_uri



import sys
import json
from fastapi.testclient import TestClient
import pytest

sys.path.append(".")

engines = ["duckdb", "polars"]


@pytest.mark.parametrize("engine", engines)
def test_nearby(client: TestClient, engine):
    response = client.post(
        f"/api/v1/test/fake_delta?limit=50&format=ndjson&%24engine={engine}",
        json={"nearby": {"lat": 46.7, "lon": 8.6, "distance_m": 10000}},
    )
    assert response.status_code == 200
    lines = [json.loads(l) for l in response.text.split("\n") if len(l) > 0]
    assert (
        len(lines) >= 15 and len(lines) <= 40
    )  # it's a bit fuzzy since distance calc is never 100% accurate

    assert lines[0]["nearby"] <= lines[1]["nearby"]
    assert lines[1]["nearby"] <= lines[2]["nearby"]
    assert lines[2]["nearby"] <= lines[3]["nearby"]
    for item in lines:
        assert item["nearby"] <= 10000


def test_no_nearby(client: TestClient):
    for e in engines:
        response = client.post(
            f"/api/v1/test/fake_delta?limit=50&format=ndjson&%24engine={e}",
            json={},
        )
        assert response.status_code == 200
        assert response.status_code == 200
        lines = [json.loads(l) for l in response.text.split("\n") if len(l) > 0]
        assert len(lines) == 50
        assert "nearby" not in lines[0]



from fastapi import FastAPI, Request, status
import dataclasses
from faker import Faker
from fastapi.exceptions import RequestValidationError
from fastapi.responses import JSONResponse
import random
import os
import logging


def get_app(default_engine="duckdb"):
    import bmsdna.lakeapi

    os.environ["MY_SQL_PWD"] = "MyPass@word4tests"
    app = FastAPI()
    def_cfg = bmsdna.lakeapi.get_default_config()
    cfg = dataclasses.replace(
        def_cfg,
        enable_sql_endpoint=True,
        data_path="tests/data",
        default_engine=default_engine,
    )

    bmsdna.lakeapi.init_lakeapi(app, True, cfg, "config_test.yml")

    @app.exception_handler(RequestValidationError)
    async def validation_exception_handler(
        request: Request, exc: RequestValidationError
    ):
        exc_str = f"{exc}".replace("\n", " ").replace("   ", " ")
        logging.error(f"{request}: {exc_str}")
        content = {"status_code": 10422, "message": exc_str, "data": None}
        return JSONResponse(
            content=content, status_code=status.HTTP_422_UNPROCESSABLE_ENTITY
        )

    @app.get("/")
    async def root(req: Request):
        return {"User": req.user["username"]}

    return app


def get_auth():
    user = "test"
    pw = "B~C:BB*_9-1u"
    return (user, pw)


lat1 = 46.7079055
lon1 = 8.6527287

lat2 = 45.752338
lon2 = 11.790724


def create_rows_faker(num=1):
    fake = Faker()
    output = [
        {
            "name": fake.name(),
            "address": fake.address(),
            "email": fake.email(),
            "bs": fake.bs(),
            "city": fake.city(),
            "state": fake.state(),
            "date_time": fake.date_time(),
            "paragraph": fake.paragraph(),
            "Conrad": fake.catch_phrase(),
            "randomdata": random.randint(1000, 2000),
            "abc": random.choice(["a", "b", "c"]),
            "geo_lat": lat1
            + (lat1 - lat2) / num * x,  # not really random, but good to test
            "geo_lon": lon1 + (lon2 - lon1) / num * x,
        }
        for x in range(num)
    ]
    return output



from typing import Sequence

from fastapi import FastAPI, Request, Response
from fastapi.security import HTTPBasic

from bmsdna.lakeapi.core.config import BasicConfig, UserConfig
from functools import lru_cache

security = HTTPBasic()
userhashmap: dict[str, str] | None = None


@lru_cache
def is_correct(
    hash: str,
    pwd_str: str,
):
    import argon2

    ph = argon2.PasswordHasher()
    return ph.verify(
        hash.encode("utf-8"),
        pwd_str.encode("utf-8"),
    )


def get_basic_auth_middleware_func(users: Sequence[UserConfig]):
    async def basic_auth_middleware_func(request: Request, call_next):
        import json

        credentials = await HTTPBasic(auto_error=False)(request)
        if credentials is None:
            return Response(
                status_code=401,
                headers={"WWW-Authenticate": "Basic"},
                content=json.dumps({"detail": "Not authenticated"}),
            )

        global userhashmap
        userhashmap = userhashmap or {
            ud["name"].casefold(): ud["passwordhash"] for ud in users if ud["name"]
        }  # pay attention not to include an empty user by accident

        if credentials.username.casefold() not in userhashmap.keys():
            return Response(
                status_code=401,
                headers={"WWW-Authenticate": "Basic"},
                content=json.dumps({"detail": "Incorrect email or password"}),
            )
        pwd_str = credentials.password
        hash = userhashmap[credentials.username.casefold()]

        is_correct_password = is_correct(hash, pwd_str)
        if not isinstance(is_correct_password, bool):
            is_correct_password = is_correct_password
        if not is_correct_password:
            return Response(
                status_code=401,
                headers={"WWW-Authenticate": "Basic"},
                content="Incorrect email or password",
            )
        request.scope["user"] = {"username": credentials.username}
        return await call_next(request)

    return basic_auth_middleware_func


def add_user_middlware(
    app: FastAPI,
    basic_config: BasicConfig,
    users: Sequence[UserConfig],
):
    from starlette.middleware.base import BaseHTTPMiddleware

    app.add_middleware(
        BaseHTTPMiddleware, dispatch=get_basic_auth_middleware_func(users)
    )



from bmsdna.lakeapi.core.config import Configs, get_default_config
import dataclasses
from bmsdna.lakeapi.core.route import init_routes
import os

def_cfg = get_default_config()  # Get default startup config
start_config = dataclasses.replace(
    def_cfg, enable_sql_endpoint=True, data_path="tests/data"
)
real_config = Configs.from_yamls(
    start_config, os.getenv("CONFIG_PATH", "config_test.yml")
)
router = init_routes(real_config, start_config)
print("done")



from fastapi.testclient import TestClient
import sys

sys.path.append(".")
import pytest

min_rounds = 1


def benchmark_engine(client: TestClient, engine="duckdb"):
    total_time = 0
    response = client.get(f"/api/v1/test/fake_{engine}?limit=10&abc=a")

    assert response.status_code == 200
    assert len(response.json()) == 10

    client.get(f"/api/v1/test/fake_delta?limit=100&abc=a&%24engine={engine}")
    client.get(f"/api/v1/test/fake_delta?limit=100&abc=b&%24engine={engine}")
    client.get(f"/api/v1/test/fake_delta?limit=100&abc=c&%24engine={engine}")

    return True


@pytest.mark.benchmark(min_rounds=min_rounds, warmup=False)
def test_benchmark_duckdb(benchmark, client: TestClient):
    result = benchmark(benchmark_engine, client, engine="duckdb")
    assert result == True


@pytest.mark.benchmark(min_rounds=min_rounds, warmup=False)
def test_benchmark_polars(benchmark, client: TestClient):
    result = benchmark(benchmark_engine, client, engine="polars")
    assert result == True



import sys
import pyarrow as pa
import pytest

sys.path.append(".")

engines = ["duckdb", "polars"]


from fastapi.testclient import TestClient


@pytest.mark.parametrize("engine", engines)
def test_data_csv4excel(client: TestClient, engine):
    # csv 4 excel is a really ... strange... format
    response = client.get(
        f"/api/v1/test/fruits?limit=1&format=csv4excel&cars=audi&%24engine={engine}"
    )
    assert response.status_code == 200

    import csv

    firstline = response.content[0:6]
    assert firstline.decode("ascii") == "sep=,\n"
    rest = response.content[6:].decode("utf-16-le")
    reader = csv.DictReader(rest.splitlines())
    line1 = reader.__next__()
    assert line1 == {"A": "2", "fruits": "banana", "B": "4", "cars": "audi"}


@pytest.mark.parametrize("engine", engines)
def test_data_csv_custom(client: TestClient, engine):
    response = client.get(
        f"/api/v1/test/fruits?limit=1&format=csv&cars=audi&%24engine={engine}&$encoding=utf-16-be&$csv_separator=|"
    )
    assert response.status_code == 200

    rest = response.content.decode("utf-16-be")
    assert "|" in rest

    lines = rest.replace("\r\n", "\n").split("\n")
    header = lines[0].replace('"', "").split("|")
    line1 = lines[1].replace('"', "").split("|")
    line1_dict = dict(zip(header, line1))

    assert line1_dict == {"A": "2", "fruits": "banana", "B": "4", "cars": "audi"}

    response = client.get(
        f"/api/v1/test/fruits?limit=1&format=csv&cars=audi&%24engine={engine}&$encoding=cp850&$csv_separator=\\t"
    )
    assert response.status_code == 200

    rest = response.content.decode("cp850")
    assert "\t" in rest
    lines = rest.replace("\r\n", "\n").split("\n")
    header = lines[0].replace('"', "").split("\t")
    line1 = lines[1].replace('"', "").split("\t")
    line1_dict = dict(zip(header, line1))
    assert line1_dict == {"A": "2", "fruits": "banana", "B": "4", "cars": "audi"}


def test_data_html(client: TestClient):
    for e in engines:
        # csv 4 excel is a really ... strange... format
        response = client.get(
            f"/api/v1/test/fruits?limit=1&format=html&cars=audi&%24engine={e}"
        )
        assert response.status_code == 200
        assert response.text.startswith("<")


def test_data_xml(client: TestClient):
    for e in engines:
        # csv 4 excel is a really ... strange... format
        response = client.get(
            f"/api/v1/test/fruits?limit=1&format=xml&cars=audi&%24engine={e}"
        )
        assert response.status_code == 200
        assert response.text.startswith("<")


def test_data_arrow_stream(client: TestClient):
    for e in engines:
        # csv 4 excel is a really ... strange... format
        response = client.get(
            f"/api/v1/test/fruits?limit=1&format=arrow-stream&cars=audi&%24engine={e}"
        )
        assert response.status_code == 200
        import tempfile

        temp_fn = tempfile.mktemp()
        with open(temp_fn, "wb") as f:
            f.write(response.content)
        with pa.OSFile(temp_fn, "rb") as fl:
            with pa.ipc.open_stream(fl) as reader:
                df = reader.read_pandas()
                assert df["A"][0] == 2



from fastapi.testclient import TestClient
import pytest
import os
from dotenv import load_dotenv


load_dotenv()


@pytest.fixture(scope="session")
def app():
    from tests.utils import get_app

    app = get_app()
    yield app


@pytest.fixture(scope="session")
def client_no_auth(app):
    client = TestClient(app)
    yield client
    client.close()


@pytest.fixture(scope="session")
def client(app):
    from tests.utils import get_auth

    client = TestClient(app)
    client.auth = get_auth()
    yield client
    client.close()


@pytest.fixture(scope="session", autouse=True)
def spawn_sql():
    import test_server

    if os.getenv("NO_SQL_SERVER", "0") == "1":
        yield None
    else:
        sql_server = test_server.start_mssql_server()
        yield sql_server
        if os.getenv("KEEP_SQL_SERVER", "0") == "0":  # can be handy during development
            sql_server.stop()


@pytest.fixture(scope="session", autouse=True)
def spawn_azurite():
    import test_server

    if os.getenv("NO_AZURITE_DOCKER", "0") == "1":
        yield None
    else:
        azurite = test_server.start_azurite()
        yield azurite
        if (
            os.getenv("KEEP_AZURITE_DOCKER", "0") == "0"
        ):  # can be handy during development
            azurite.stop()



import requests

user = "test"
pw = "B~C:BB*_9-1u"
auth = (user, pw)
engines = ["polars"]
client = requests.Session()
response = client.get("http://localhost:8080/metadata", auth=auth)
assert response.status_code == 200
jsd = response.json()
for item in jsd:
    for e in engines:
        name = item["name"]
        tag = item["tag"]
        route = item["route"]
        meta_detail_route = (
            "http://localhost:8080/" + route + f"/metadata_detail?%24engine={e}"
        )
        print(meta_detail_route)
        response = client.get(meta_detail_route, auth=auth)
        if name not in ["not_existing", "not_existing2"]:
            assert name + "_" + str(response.status_code) == name + "_200"
        else:
            assert name + "_" + str(response.status_code) == name + "_404"



from fastapi.testclient import TestClient
import sys
import pytest

sys.path.append(".")

engines = ("duckdb", "polars")


@pytest.mark.parametrize("engine", engines)
def test_data_partition(client: TestClient, engine):
    for _ in range(2):
        response = client.get(
            f"/api/v1/test/fruits_partition?limit=1&format=json&cars=audi&%24engine={engine}"
        )
        assert response.status_code == 200
        assert response.json() == [
            {
                "A": 2,
                "fruits": "banana",
                "B": 4,
                "cars": "audi",
                "my_empty_col": None,
            }
        ]
        response = client.get(
            f"/api/v1/test/fruits_partition?limit=1&format=json&fruits=ananas&%24engine={engine}"
        )
        assert response.status_code == 200
        assert response.json() == [
            {
                "A": 9,
                "fruits": "ananas",
                "B": 9,
                "cars": "fiat",
                "my_empty_col": None,
            }
        ]


@pytest.mark.parametrize("engine", engines)
def test_data_partition_mod(client: TestClient, engine):
    for _ in range(2):
        response = client.get(
            f"/api/v1/test/fruits_partition_mod?limit=1&format=json&cars=audi&%24engine={engine}"
        )
        assert response.status_code == 200
        assert response.json() == [
            {
                "A": 2,
                "fruits": "banana",
                "B": 4,
                "cars": "audi",
            }
        ]
        response = client.post(
            f"/api/v1/test/fruits_partition_mod?limit=1&format=json&%24engine={engine}",
            json={"cars_in": ["audi"]},
        )
        assert response.status_code == 200
        assert response.json() == [
            {
                "A": 2,
                "fruits": "banana",
                "B": 4,
                "cars": "audi",
            }
        ]


@pytest.mark.parametrize("engine", engines)
def test_data_partition_int(client: TestClient, engine):
    for _ in range(2):
        response = client.get(
            f"/api/v1/test/fruits_partition_int?limit=1&format=json&A=2&%24engine={engine}"
        )
        assert response.status_code == 200
        assert response.json() == [
            {
                "A": 2,
                "fruits": "banana",
                "B": 4,
                "cars": "audi",
            }
        ]



import sys
from fastapi.testclient import TestClient

sys.path.append(".")

engines = ["duckdb", "polars"]


def test_returns_complex(client: TestClient):
    for e in engines:
        response = client.post(
            f"/api/v1/complexer/complex_fruits?limit=10&format=json&%24engine={e}",
        )
        assert response.status_code == 200
        jsd = response.json()
        assert len(jsd) == 6
        assert isinstance(jsd[0]["vitamines"], list)
        assert isinstance(jsd[0]["person"], dict)


def test_returns_jsonify(client: TestClient):
    for e in engines:
        response = client.post(
            f"/api/v1/complexer/complex_fruits?limit=10&format=json&jsonify_complex=True&%24engine={e}",
        )
        assert response.status_code == 200
        jsd = response.json()
        assert len(jsd) == 6
        assert isinstance(jsd[0]["vitamines"], str)
        assert isinstance(jsd[0]["person"], str)
        import json

        assert isinstance(json.loads(jsd[0]["vitamines"]), list)
        assert isinstance(json.loads(jsd[0]["person"]), dict)


def test_returns_metadatadeta(client: TestClient):
    response = client.get(
        "/api/v1/complexer/complex_fruits/metadata_detail",
    )
    assert response.status_code == 200

    jsd = response.json()

    vit = [p for p in jsd["data_schema"] if p["name"] == "vitamines"][0]
    assert vit is not None
    assert vit["type"]["type_str"].lower().startswith("list<")

    per = [p for p in jsd["data_schema"] if p["name"] == "person"][0]
    assert per is not None
    assert per["type"]["type_str"].lower().startswith("struct<")

    assert "person" not in jsd["max_string_lengths"]
    assert "vitamines" not in jsd["max_string_lengths"]
    assert "fruits" in jsd["max_string_lengths"]


def test_returns_metadatadeta_partition(client: TestClient):
    response = client.get(
        "/api/v1/test/fruits_partition/metadata_detail",
    )
    assert response.status_code == 200

    jsd = response.json()

    assert jsd["max_string_lengths"]["my_empty_col"] is None


def test_returns_metadatadeta_jsonifiyed(client: TestClient):
    response = client.get(
        "/api/v1/complexer/complex_fruits/metadata_detail?jsonify_complex=True",
    )
    assert response.status_code == 200

    jsd = response.json()

    vit = [p for p in jsd["data_schema"] if p["name"] == "vitamines"][0]
    assert vit is not None
    assert vit["type"]["type_str"].lower() == "string"

    per = [p for p in jsd["data_schema"] if p["name"] == "person"][0]
    assert per is not None
    assert per["type"]["type_str"].lower() == "string"

    assert "person" in jsd["max_string_lengths"]
    assert "vitamines" in jsd["max_string_lengths"]
    assert "fruits" in jsd["max_string_lengths"]


def test_returns_csv(client: TestClient):
    for e in engines:
        response = client.post(
            f"/api/v1/complexer/complex_fruits?limit=10&format=csv&%24engine={e}",
        )
        assert response.status_code == 200
        import csv

        reader = csv.DictReader(response.text.splitlines())
        line1 = reader.__next__()
        assert isinstance(line1["vitamines"], str)
        assert isinstance(line1["person"], str)
        import json

        assert isinstance(json.loads(line1["vitamines"]), list)
        assert isinstance(json.loads(line1["person"]), dict)



import sys
from fastapi.testclient import TestClient

sys.path.append(".")

engines = ["duckdb", "polars"]


def test_not_in(client: TestClient):
    for e in engines:
        response = client.post(
            f"/api/v1/complexer/complex_fruits?limit=5&format=json&%24engine={e}",
            json={"cars_not_in": ["audi", "fiat"]},
        )
        assert response.status_code == 200
        jsd = response.json()
        assert len(jsd) > 0

        for item in jsd:
            assert item["cars"] != "audi"
            assert item["cars"] != "fiat"


def test_in(client: TestClient):
    for e in engines:
        response = client.post(
            f"/api/v1/complexer/complex_fruits?limit=5&format=json&%24engine={e}",
            json={"fruits_in": ["banana", "kiwi"]},
        )
        assert response.status_code == 200
        jsd = response.json()
        assert len(jsd) > 0

        for item in jsd:
            assert item["fruits"] == "banana"


def test_contains(client: TestClient):
    for e in engines:
        response = client.post(
            f"/api/v1/complexer/complex_fruits?limit=5&format=json&%24engine={e}",
            json={"fruits_contains": "anan"},
        )
        assert response.status_code == 200
        jsd = response.json()
        assert len(jsd) > 0

        for item in jsd:
            assert item["fruits"] in ["banana", "ananas"]


def test_startswith(client: TestClient):
    for e in engines:
        response = client.post(
            f"/api/v1/complexer/complex_fruits?limit=5&format=json&%24engine={e}",
            json={"fruits_startswith": "anan"},
        )
        assert response.status_code == 200
        jsd = response.json()
        assert len(jsd) > 0

        for item in jsd:
            assert item["fruits"] in ["ananas"]
            assert item["fruits"] not in ["banana"]


def test_not_contains(client: TestClient):
    for e in engines:
        response = client.post(
            f"/api/v1/complexer/complex_fruits?limit=5&format=json&%24engine={e}",
            json={"fruits_not_contains": "anan"},
        )
        assert response.status_code == 200
        jsd = response.json()
        assert len(jsd) > 0

        for item in jsd:
            assert item["fruits"] not in ["banana", "ananas"]


def test_not_equals(client: TestClient):
    for e in engines:
        response = client.post(
            f"/api/v1/complexer/complex_fruits?limit=5&format=json&%24engine={e}",
            json={"fruits_ne": "banana"},
        )
        assert response.status_code == 200
        jsd = response.json()
        assert len(jsd) > 0

        for item in jsd:
            assert item["fruits"] != "banana"


def test_gt(client: TestClient):
    for e in engines:
        response = client.post(
            f"/api/v1/complexer/complex_fruits?limit=5&format=json&%24engine={e}",
            json={"B_gt": "2"},
        )
        assert response.status_code == 200
        jsd = response.json()
        assert len(jsd) > 0

        for item in jsd:
            assert item["B"] > 2


def test_gt_date_get(client: TestClient):
    from datetime import datetime

    for e in ["duckdb"]:  # TODO: Enable polalrs once they support datetime decently
        response = client.get(
            f"/api/v1/test/fruits_date?limit=5&format=json&%24engine={e}&date_field_gt=2023-01-01T00:00",
        )
        assert response.status_code == 200
        jsd = response.json()
        assert len(jsd) > 0

        for item in jsd:
            assert datetime.fromisoformat(item["date_field"]) > datetime(2023, 1, 1)


def test_gte(client: TestClient):
    for e in engines:
        response = client.post(
            f"/api/v1/complexer/complex_fruits?limit=5&format=json&%24engine={e}",
            json={"B_gte": "2"},
        )
        assert response.status_code == 200
        jsd = response.json()
        assert len(jsd) > 0

        for item in jsd:
            assert item["B"] >= 2


def test_lt(client: TestClient):
    for e in engines:
        response = client.post(
            f"/api/v1/complexer/complex_fruits?limit=5&format=json&%24engine={e}",
            json={"B_lt": 7},
        )
        assert response.status_code == 200
        jsd = response.json()
        assert len(jsd) > 0

        for item in jsd:
            assert item["B"] < 7


def test_lte(client: TestClient):
    for e in engines:
        response = client.post(
            f"/api/v1/complexer/complex_fruits?limit=5&format=json&%24engine={e}",
            json={"B_lte": 7},
        )
        assert response.status_code == 200
        jsd = response.json()
        assert len(jsd) > 0

        for item in jsd:
            assert item["B"] <= 7


def test_between(client: TestClient):
    for e in engines:
        response = client.post(
            f"/api/v1/complexer/complex_fruits?limit=5&format=json&%24engine={e}",
            json={"B_between": [5, 7]},
        )
        assert response.status_code == 200
        jsd = response.json()
        assert len(jsd) > 0

        for item in jsd:
            assert item["B"] >= 5 and item["B"] <= 7

        response = client.post(
            f"/api/v1/complexer/complex_fruits?limit=5&format=json&%24engine={e}",
            json={"B_between": [5, 7, 9]},
        )
        assert response.status_code == 400


def test_not_between(client: TestClient):
    for e in engines:
        response = client.post(
            f"/api/v1/complexer/complex_fruits?limit=5&format=json&%24engine={e}",
            json={"B_not_between": [5, 7]},
        )
        assert response.status_code == 200
        jsd = response.json()
        assert len(jsd) > 0

        for item in jsd:
            assert not (item["B"] >= 5 and item["B"] <= 7)

        response = client.post(
            f"/api/v1/complexer/complex_fruits?limit=5&format=json&%24engine={e}",
            json={"B_not_between": [5, 7, 9]},
        )
        assert response.status_code == 400


def test_has(client: TestClient):
    for e in engines:
        response = client.get(
            f"/api/v1/array/weather?limit=100&temperatures_has=E1&format=json&%24engine={e}",
        )
        assert response.status_code == 200
        jsd = response.json()
        assert len(jsd) == 3

        for item in jsd:
            assert "E1" in item["temperatures"]

        response = client.post(
            f"/api/v1/array/weather?limit=100&format=json&%24engine={e}",
            json={"temperatures_has": "E1"},
        )
        assert response.status_code == 200
        jsd = response.json()
        assert len(jsd) == 3

        for item in jsd:
            assert "E1" in item["temperatures"]



from bmsdna.lakeapi.core.datasource import Datasource
from bmsdna.lakeapi.core.config import BasicConfig
from bmsdna.lakeapi.utils.async_utils import _async


def get_schema_cached(cfg: BasicConfig, datasource: Datasource, key: str):
    if cfg.schema_cache_ttl is not None:
        import os
        import time
        import pyarrow.parquet as pq

        schema_cache_file = os.path.join(
            cfg.temp_folder_path,
            key + ".parquet",
        )
        if (
            not os.path.exists(schema_cache_file)
            or (time.time() - os.path.getmtime(schema_cache_file))
            > cfg.schema_cache_ttl
        ):
            if not datasource.file_exists():
                schema = None
            else:
                schema = datasource.get_schema()
                pq.write_metadata(schema, schema_cache_file)
        else:
            schema = pq.read_schema(schema_cache_file)
        return schema
    else:
        if not datasource.file_exists():
            return None
        return datasource.get_schema()



from pathlib import Path
import docker
from docker.models.containers import Container
from time import sleep
from typing import cast
import docker.errors
import os
from azure.core.exceptions import ResourceExistsError


def _getenvs():
    envs = dict()
    with open("test_server/sql_docker.env", "r") as f:
        lines = f.readlines()
        envs = {
            item[0].strip(): item[1].strip()
            for item in [
                line.split("=")
                for line in lines
                if len(line.strip()) > 0 and not line.startswith("#")
            ]
        }
    return envs


def start_mssql_server() -> Container:
    client = docker.from_env()  # code taken from https://github.com/fsspec/adlfs/blob/main/adlfs/tests/conftest.py#L72
    sql_server: Container | None = None
    try:
        m = cast(Container, client.containers.get("test4sql"))
        if m.status == "running":
            return m
        else:
            sql_server = m
    except docker.errors.NotFound:
        pass

    envs = _getenvs()

    if sql_server is None:
        # using podman:  podman run  --env-file=TESTS/SQL_DOCKER.ENV --publish=1439:1433 --name=mssql1 chriseaton/adventureworks:light
        #                podman kill mssql1
        sql_server = client.containers.run(
            "chriseaton/adventureworks:light",
            environment=envs,
            detach=True,
            name="test4sql",
            ports={"1433/tcp": 1439},
        )  # type: ignore
    assert sql_server is not None
    sql_server.start()
    print(sql_server.status)
    sleep(
        45
    )  # the install script takes a sleep of 30s to create the db and then restores adventureworks
    import pyodbc

    with pyodbc.connect(
        "DRIVER={ODBC Driver 18 for SQL Server};SERVER=127.0.0.1,1439;ENCRYPT=yes;TrustServerCertificate=Yes;UID=sa;PWD="
        + envs["MSSQL_SA_PASSWORD"]
        + ";DATABASE=master"
    ) as con:
        with con.cursor() as cur:
            cur.execute("SELECT * FROM sys.databases where name='AdventureWorks'")
            if len(cur.fetchall()) > 0:
                print("db created")
            else:
                print("db not created yet")
                sleep(25)
    print("Successfully created sql container...")
    return sql_server


def get_test_blobstorage():
    constr = os.getenv(
        "TEST_BLOB_CONSTR",
        "DefaultEndpointsProtocol=http;AccountName=devstoreaccount1;AccountKey=Eby8vdM02xNOcqFlqUwJPLlmEtlCDXJ1OUzFT50uSRZ6IFsuFq2UVErCz4I6tq/K1SZFPTOtr/KBHBeksoGMGw==;BlobEndpoint=http://127.0.0.1:10000/devstoreaccount1;",
    )
    from azure.storage.blob import ContainerClient

    cc = ContainerClient.from_connection_string(constr, "testlake")
    if not cc.exists():
        cc.create_container()
    return cc


def upload_to_azurite():
    with get_test_blobstorage() as cc:
        faker_pq = "tests/data/startest/faker.parquet"
        try:
            with open(faker_pq, "rb") as f:
                cc.upload_blob("td/faker.parquet", f)
        except ResourceExistsError:
            pass  # already uploaded
        fakeroot = Path("tests/data/delta/fake")
        for root, _, fls in os.walk(fakeroot):
            for fl in fls:
                try:
                    rel = str(Path(root).relative_to(fakeroot))
                    with open(os.path.join(root, fl), "rb") as f:
                        target_path = (
                            f"td/delta/fake/{rel}/{fl}"
                            if rel != "."
                            else f"td/delta/fake/{fl}"
                        )
                        cc.upload_blob(target_path, f)
                except ResourceExistsError:
                    pass  # already uploaded


def start_azurite() -> Container:
    client = docker.from_env()  # code taken from https://github.com/fsspec/adlfs/blob/main/adlfs/tests/conftest.py#L72
    azurite_server: Container | None = None
    try:
        m = cast(Container, client.containers.get("test4azurite"))
        if m.status == "running":
            upload_to_azurite()
            return m
        else:
            azurite_server = m
    except docker.errors.NotFound:
        pass

    if azurite_server is None:
        azurite_server = client.containers.run(
            "mcr.microsoft.com/azure-storage/azurite:latest",
            detach=True,
            name="test4azurite",
            ports={"10000/tcp": 10000, "10001/tcp": 10001, "10002/tcp": 10002},
        )  # type: ignore
    assert azurite_server is not None
    azurite_server.start()
    print(azurite_server.status)
    sleep(20)
    upload_to_azurite()
    print("Successfully created azurite container...")
    return azurite_server



from fastapi.testclient import TestClient

engines = ["duckdb", "polars"]


def test_tables(client: TestClient):
    for e in engines:
        response = client.get(f"/api/sql/tables?%24engine={e}")
        assert response.status_code == 200
        tables = response.json()
        assert len(tables) > 5


def test_get(client: TestClient):
    for e in engines:
        response = client.get(
            f"/api/sql?%24engine={e}&sql=SELECT distinct B FROM complexer_complex_fruits union select distinct A as B FROM startest_fruits"
        )
        assert response.status_code == 200
        tables = response.json()
        assert len(tables) > 5


def test_post(client: TestClient):
    for e in engines:
        response = client.post(
            f"/api/sql?%24engine={e}&",
            data="SELECT distinct B FROM complexer_complex_fruits union select distinct  A as B FROM startest_fruits",  # type: ignore
        )
        assert response.status_code == 200
        tables = response.json()
        assert len(tables) > 5


def test_sql_where_post(client: TestClient):
    # better naming needed in the future

    query = """select * 
        from test_fruits 
        where cars = 'audi' 
        and fruits = 'banana' 
        and A = 2 and B = 4"""

    response = client.post(
        "/api/sql",
        data=query,  # type: ignore
    )
    assert response.status_code == 200
    print(response.json())
    assert response.json() == [
        {
            "A": 2,
            "fruits": "banana",
            "B": 4,
            "cars": "audi",
        }
    ]


def test_sql_where_get(client: TestClient):
    # better naming needed in the future

    query = """select * 
        from test_fruits 
        where cars = 'audi' 
        and fruits = 'banana' 
        and A = 2 and B = 4"""

    response = client.get(
        "/api/sql",
        params={"sql": query},
    )
    assert response.status_code == 200
    print(response.json())
    assert response.json() == [
        {
            "A": 2,
            "fruits": "banana",
            "B": 4,
            "cars": "audi",
        }
    ]



import inspect
from typing import Awaitable, TypeVar

T = TypeVar("T")


async def _async(thing: T | Awaitable[T]) -> T:
    if inspect.isawaitable(thing):
        return await thing
    return thing



from typing import List, TYPE_CHECKING
from bmsdna.lakeapi.context.source_uri import SourceUri
from deltalake2db.delta_meta_retrieval import (
    get_meta,
    PolarsEngine,
)
import logging

if TYPE_CHECKING:
    from bmsdna.lakeapi.core.types import Param
    from bmsdna.lakeapi.core.config import BasicConfig

logger = logging.getLogger(__name__)


def _with_implicit_parameters(
    paramslist: "List[Param]",
    file_type: str,
    uri: SourceUri,
    basic_config: "BasicConfig",
):
    if file_type == "delta":
        fs, fs_spec = uri.get_fs_spec()
        if not fs.exists(fs_spec + "/_delta_log"):
            return paramslist

        try:
            dt_uri, dt_opts = uri.get_uri_options(flavor="object_store")
            meta = get_meta(PolarsEngine(dt_opts), dt_uri)  # to ensure connectivity
            assert meta.last_metadata is not None
            part_cols = meta.last_metadata.get("partitionColumns", [])
            if part_cols and len(part_cols) > 0:
                all_names = [(p.colname or p.name).lower() for p in paramslist]
                new_params = list(paramslist)
                for pc in part_cols:
                    if (
                        pc.lower() not in all_names
                        and not basic_config.should_hide_col_name(pc)
                    ):
                        from bmsdna.lakeapi.core.types import Param

                        new_params.append(Param(pc, operators=["="], colname=pc))
                return new_params
        except FileNotFoundError:
            return paramslist  # this is not critical here
        except Exception as e:
            logger.warning(f"Error retrieving delta metadata: {e}")
            return paramslist  # this is not critical here

    return paramslist



from typing import List, Literal, Optional, Type, Union, cast, Any

import pyarrow as pa


import sqlglot.expressions as ex

from fastapi import APIRouter, Depends, Header, HTTPException, Query, Request
from pydantic import BaseModel

from bmsdna.lakeapi.context import get_context_by_engine
from bmsdna.lakeapi.context.source_uri import SourceUri
from bmsdna.lakeapi.context.df_base import ExecutionContext, get_sql
from bmsdna.lakeapi.core.config import BasicConfig, Config, Configs
from bmsdna.lakeapi.core.datasource import (
    Datasource,
    filter_df_based_on_params,
    get_filters,
    filter_partitions_based_on_params,
)
from bmsdna.lakeapi.core.log import get_logger
from bmsdna.lakeapi.core.model import create_parameter_model, create_response_model
from bmsdna.lakeapi.core.response import create_response
from bmsdna.lakeapi.core.types import Engines, OperatorType, OutputFileType
from bmsdna.lakeapi.endpoint.endpoint_search import handle_search_request
from bmsdna.lakeapi.endpoint.endpoint_nearby import handle_nearby_request
from starlette.concurrency import run_in_threadpool
from bmsdna.lakeapi.utils.async_utils import _async
from deltalake2db.delta_meta_retrieval import (
    PolarsEngine,
    get_meta,
)


logger = get_logger(__name__)


def remove_search_nearby(
    prm_dict: dict,
    config: Config,
):
    if (not config.search and not config.nearby) or len(prm_dict) == 0:
        return prm_dict
    search_nearby_cols = [s.name.lower() for s in (config.search or [])] + [
        s.name.lower() for s in (config.nearby or [])
    ]
    return {k: v for k, v in prm_dict.items() if k.lower() not in search_nearby_cols}


def get_params_filter_expr(
    context: ExecutionContext,
    columns: List[str],
    config: Config,
    params: BaseModel,
) -> Optional[ex.Condition | ex.Binary]:
    expr = filter_df_based_on_params(
        context,
        remove_search_nearby(
            params.model_dump(exclude_unset=True) if params else {}, config
        ),
        config.params if config.params else [],
        columns,
    )
    return expr


def get_response_model(
    config: Config,
    schema: pa.Schema,
    basic_config: BasicConfig,
) -> Optional[Type[BaseModel]]:
    response_model: Optional[type[BaseModel]] = None
    try:
        response_model = create_response_model(
            config.tag + "_" + config.name, schema, basic_config=basic_config
        )
    except Exception as err:
        logger.warning(f"Could not get response type for f{config.route}. Error:{err}")
        response_model = None
    return response_model


def exclude_cols(columns: List[str], config: BasicConfig) -> List[str]:
    columns = [c for c in columns if not config.should_hide_col_name(c)]
    return columns


def split_csv(csv_str: str) -> list[str]:
    import csv

    reader = csv.reader([csv_str], delimiter=",", quotechar='"')
    for i in reader:
        return i
    raise ValueError("cannot happen")


def is_json_response(
    result,
    args,
    kwargs,
    key=None,
):
    return (
        kwargs.get("format") == "json"
        or kwargs.get("format") == "ndjson"
        or kwargs.get("request").headers.get("Accept") == "application/json"
        or kwargs.get("request").headers.get("Accept") == "application/x-ndjson"
    )


def is_complex_type(
    schema: pa.Schema,
    col_name: str,
):
    f = schema.field(col_name)
    return pa.types.is_nested(f.type)


def create_config_endpoint(
    schema: pa.Schema | None,
    apimethod: Literal["get", "post"],
    config: Config,
    router: APIRouter,
    response_model: Type[BaseModel] | None,
    basic_config: BasicConfig,
    configs: Configs,
):
    route = config.route

    query_model = create_parameter_model(
        schema,
        config.tag + "_" + config.name + "_" + apimethod,
        config.params,
        config.search,
        config.nearby,
        apimethod,
    )

    api_method_mapping = {
        "post": router.post(
            route,
            tags=[config.tag],
            response_model=response_model,
            operation_id=config.tag + "_" + config.name + "_" + apimethod,
            name=config.name,
        ),
        "get": router.get(
            route,
            tags=[config.tag],
            response_model=response_model,
            operation_id=config.tag + "_" + config.name + "_" + apimethod,
            name=config.name,
        ),
    }

    api_method = api_method_mapping[apimethod]
    has_complex = True
    if schema is not None:
        has_complex = any(
            (pa.types.is_struct(t) or pa.types.is_list(t) for t in schema.types)
        )

    @api_method
    async def data(
        request: Request,
        params: query_model = (Depends() if apimethod == "get" else None),  # type: ignore
        Accept: Union[str, None] = Header(default=None),
        limit: Optional[int] = 100,
        offset: Optional[int] = 0,
        select: Union[str, None] = Query(
            title="$select",
            alias="$select",
            default=None,
            include_in_schema=False,
        ),
        distinct: bool = Query(
            title="$distinct",
            alias="$distinct",
            default=False,
            include_in_schema=False,
        ),
        engine: Engines | None = Query(
            title="$engine",
            alias="$engine",
            default=None,
            include_in_schema=False,
        ),
        format: Optional[OutputFileType] = "json",
        jsonify_complex: bool = Query(
            title="jsonify_complex",
            include_in_schema=has_complex,
            default=False,
        ),
        chunk_size: int | None = Query(
            title="$chunk_size",
            include_in_schema=False,
            default=None,
        ),
    ):  # type: ignore
        logger.debug(
            f"{params.model_dump(exclude_unset=True) if params else None}Union[ ,  ]{request.url.path}"
        )

        engine = engine or config.engine or basic_config.default_engine

        logger.debug(f"Engine: {engine}")
        real_chunk_size = (
            chunk_size or config.chunk_size or basic_config.default_chunk_size
        )
        context = get_context_by_engine(
            engine,
            chunk_size=real_chunk_size,
        )
        assert config.datasource is not None
        with Datasource(
            config.version_str,
            config.tag,
            config.name,
            config=config.datasource,
            sql_context=context,
            basic_config=basic_config,
            accounts=configs.accounts,
        ) as realdataframe:
            if params:
                pre_filter, _ = get_filters(
                    params.model_dump(exclude_unset=True),
                    config.params if config.params else [],
                    None,
                )
            else:
                pre_filter = None
            df = realdataframe.get_df(filters=pre_filter)
            df_cols = df.columns()
            expr = get_params_filter_expr(  # this supports all kinds of filters, while the prefilter only supports equality
                context,
                df_cols,
                config,
                params,
            )
            new_query = df.query_builder()
            new_query = new_query.where(expr) if expr is not None else new_query
            columns = exclude_cols(df_cols, basic_config)
            if select:
                columns = [
                    c for c in columns if c in split_csv(select)
                ]  # split , is a bit naive, we might want to support real CSV with quotes here
            if config.datasource.exclude and len(config.datasource.exclude) > 0:
                columns = [c for c in columns if c not in config.datasource.exclude]
            if config.datasource.sortby:
                for s in config.datasource.sortby:
                    new_query = cast(ex.Select, new_query).order_by(
                        ex.column(s.by, quoted=True).desc()
                        if s.direction and s.direction.lower() == "desc"
                        else ex.column(s.by, quoted=True),
                        copy=False,
                    )
            if has_complex and format in ["csv", "excel", "scsv", "csv4excel"]:
                jsonify_complex = True
            if jsonify_complex:
                base_schema = df.arrow_schema()

                complex_cols = [c for c in columns if is_complex_type(base_schema, c)]

                new_query = context.jsonify_complex(new_query, complex_cols, columns)
            else:
                new_query = new_query.select(
                    *[ex.column(c, quoted=True) for c in columns], append=False
                )

            if distinct:
                assert len(columns) <= 3  # reduce complexity here
                new_query = cast(ex.Select, new_query).distinct()

            if not (limit == -1 and config.allow_get_all_pages):
                limit = (1000 if limit == -1 else limit) or 1000
                new_query = new_query.limit(limit)
                if offset:
                    new_query.offset(offset, copy=False)

            new_query = handle_search_request(
                context,
                config,
                params,
                basic_config,
                source_view=realdataframe.tablename,
                query=new_query,
            )
            new_query = handle_nearby_request(
                context,
                config,
                params,
                basic_config,
                source_view=realdataframe.tablename,
                query=new_query,
            )
            logger.debug(f"Query: {get_sql(new_query, dialect='duckdb')}")

            try:
                return await create_response(
                    request.url,
                    request.query_params,
                    format or request.headers["Accept"],
                    context=context,
                    sql=new_query,
                    basic_config=basic_config,
                    close_context=True,
                    charset=request.query_params.get("$encoding"),
                )
            except Exception as err:
                logger.error("Error in creating response", exc_info=err)
                raise HTTPException(status_code=500)



from bmsdna.lakeapi.context.source_uri import _convert_options
from azure.core.credentials import TokenCredential, AccessToken
from datetime import datetime


def fake_token(**kwargs):
    if len(kwargs) > 0:
        kwargs_str = ",".join([f"{k}={v}" for k, v in kwargs.items()])
        return "fake_token_" + kwargs_str
    return "fake_token"


class FakeCredential(TokenCredential):
    def __init__(self, _type):
        self._type = _type

    def get_token(self, *args, **kwargs):
        exp = int(datetime.now().timestamp()) + 72000
        if len(kwargs) > 0:
            kwargs_str = ",".join([f"{k}={v}" for k, v in kwargs.items()])
            return AccessToken(token="fake_token_" + kwargs_str, expires_on=exp)
        return AccessToken(token="fake_token", expires_on=exp)


def test_fsspec():
    o_non_anon = {"account_name": "blubb"}
    full_url = "https://blubb.blob.core.windows.net/xyz/abc"
    _, o = _convert_options(
        full_url,
        o_non_anon,
        flavor="fsspec",
        token_retrieval_func=lambda x: FakeCredential(x),
    )
    assert o == {"account_name": "blubb", "anon": False}
    assert "anon" not in o_non_anon

    o_anon = {"account_name": "blubb", "anon": True}
    _, o = _convert_options(
        full_url, o_anon, flavor="fsspec", token_retrieval_func=FakeCredential
    )
    assert o == {"account_name": "blubb", "anon": True}


def test_account_key():
    o_non_anon = {"account_name": "blubb", "account_key": "nix"}
    full_url = "https://blubb.blob.core.windows.net/xyz/abc"
    _, o = _convert_options(
        full_url, o_non_anon, flavor="object_store", token_retrieval_func=FakeCredential
    )
    assert o == {"account_name": "blubb", "account_key": "nix"}

    _, o2 = _convert_options(
        full_url, o_non_anon, flavor="fsspec", token_retrieval_func=FakeCredential
    )
    assert o2 == {"account_name": "blubb", "account_key": "nix"}


if __name__ == "__main__":
    test_fsspec()
    test_account_key()



from bmsdna.lakeapi.core.config import Configs
from bmsdna.lakeapi.api.api import init_lakeapi
from bmsdna.lakeapi.core.config import get_default_config, BasicConfig
from bmsdna.lakeapi.core.types import (
    MetadataDetailResult,
    MetadataSchemaField,
    MetadataSchemaFieldType,
)
from bmsdna.lakeapi.core.uservalidation import add_user_middlware



from fastapi.testclient import TestClient
from .utils import get_app, get_auth, create_rows_faker
import sys
import polars as pl
from deltalake import write_deltalake
import os
from hashlib import md5

sys.path.append(".")

auth = get_auth()
engines = ["duckdb", "polars"]


def test_data_overwrite(client: TestClient):
    for e in engines:
        response = client.get(
            f"/api/v1/test/fake_delta?limit=1000&&format=json&%24engine={e}"
        )
        assert response.status_code == 200
        assert len(response.json()) == 1000

    df_faker = pl.DataFrame(create_rows_faker(1001)).to_pandas()

    df_faker["name_md5_prefix_2"] = [
        md5(val.encode("UTF-8")).hexdigest()[:1] for val in df_faker["name"]
    ]

    df_faker["name1"] = df_faker["name"]

    print(df_faker)
    assert os.path.exists("tests/data/delta/fake/_delta_log")
    write_deltalake(
        "tests/data/delta/fake", df_faker, mode="overwrite", schema_mode="overwrite"
    )

    for e in engines:
        response = client.get(
            f"/api/v1/test/fake_delta?limit=1000&&format=json&%24engine={e}"
        )
        assert response.status_code == 200
        assert len(response.json()) == 1000



from datetime import datetime

from fastapi.concurrency import run_in_threadpool
import pyarrow as pa
from typing import List, Optional, Tuple, Any, Union, cast
from bmsdna.lakeapi.core.types import FileTypes, OperatorType
from bmsdna.lakeapi.context.df_base import ExecutionContext, ResultData, get_sql
from deltalake2db import (
    duckdb_create_view_for_delta,
    duckdb_apply_storage_options,
    FilterType,
)
import duckdb
import pyarrow.dataset
import sqlglot.expressions as ex
from sqlglot import from_, parse_one
import os
from datetime import timezone
from bmsdna.lakeapi.core.config import SearchConfig
from uuid import uuid4

from bmsdna.lakeapi.core.log import get_logger
import multiprocessing
from .source_uri import SourceUri


logger = get_logger(__name__)

ENABLE_COPY_TO = os.environ.get("ENABLE_COPY_TO", "0") == "1"

DUCK_CONFIG = {}
DUCK_INIT_SCRIPTS: list[str] = []
AZURE_LOADED_SCRIPTS: list[str] = []


def _get_temp_table_name():
    return "temp_" + str(uuid4()).replace("-", "")


class DuckDBResultData(ResultData):
    def __init__(
        self,
        original_sql: Union[ex.Query, str],
        con: duckdb.DuckDBPyConnection,
        chunk_size: int,
    ) -> None:
        super().__init__(chunk_size=chunk_size)
        self.original_sql = original_sql
        self.con = con
        self._arrow_schema = None
        self._df = None

    def columns(self):
        return (self.arrow_schema()).names

    def query_builder(self) -> ex.Select:
        if not isinstance(self.original_sql, str):
            return from_(self.original_sql.subquery())
        else:
            return from_(
                cast(
                    ex.Select, parse_one(self.original_sql, dialect="duckdb")
                ).subquery()
            )

    def arrow_schema(self) -> pa.Schema:
        if self._arrow_schema is not None:
            return self._arrow_schema
        query = get_sql(self.original_sql, limit=0, dialect="duckdb")

        def _get_schema():
            self.con.execute(query)
            return self.con.arrow().schema

        self._arrow_schema = _get_schema()

        return self._arrow_schema

    async def to_pandas(self):
        query = get_sql(self.original_sql, dialect="duckdb")

        def _to_df():
            self.con.execute(query)
            return self.con.df()

        return await run_in_threadpool(_to_df)

    async def to_pylist(self):
        query = get_sql(self.original_sql, dialect="duckdb")

        def _to_list():
            self.con.execute(query)
            assert self.con.description is not None
            col_names = [d[0] for d in self.con.description]
            res = self.con.fetchall()
            return [dict(zip(col_names, r)) for r in res]

        return await run_in_threadpool(_to_list)

    async def to_arrow_recordbatch(self, chunk_size: int = 10000):
        query = get_sql(self.original_sql, dialect="duckdb")

        def _to_df():
            self.con.execute(query)
            return self.con.fetch_record_batch(chunk_size)

        return await run_in_threadpool(_to_df)

    async def write_parquet(self, file_name: str):
        if not ENABLE_COPY_TO:
            return await super().write_parquet(file_name)
        query = get_sql(self.original_sql, dialect="duckdb")
        uuidstr = _get_temp_table_name()
        # temp table required because of https://github.com/duckdb/duckdb/issues/7616
        full_query = f"""CREATE TEMP VIEW {uuidstr} AS {query};
                         COPY (SELECT *FROM {uuidstr})
                         TO '{file_name}' (FORMAT PARQUET,use_tmp_file False, ROW_GROUP_SIZE 10000);
                         DROP VIEW {uuidstr}"""

        await run_in_threadpool(self.con.execute, full_query)

    async def write_nd_json(self, file_name: str):
        if not ENABLE_COPY_TO:
            return await super().write_nd_json(file_name)
        query = get_sql(self.original_sql, dialect="duckdb")
        uuidstr = _get_temp_table_name()
        full_query = f"""CREATE TEMP VIEW {uuidstr} AS {query};
                         COPY (SELECT *FROM {uuidstr})
                         TO '{file_name}' (FORMAT JSON);
                         DROP VIEW {uuidstr}"""

        await run_in_threadpool(self.con.execute, full_query)

    async def write_csv(self, file_name: str, *, separator: str):
        if not ENABLE_COPY_TO:
            return await super().write_csv(file_name, separator=separator)
        query = get_sql(self.original_sql, dialect="duckdb")
        uuidstr = _get_temp_table_name()
        full_query = f"""CREATE TEMP VIEW {uuidstr} AS {query};
                         COPY (SELECT *FROM {uuidstr}) 
                         TO '{file_name}' (FORMAT CSV, delim '{separator}', header True);
                         DROP VIEW {uuidstr};"""
        await run_in_threadpool(self.con.execute, full_query)

    async def write_json(self, file_name: str):
        if not ENABLE_COPY_TO:
            return await super().write_json(file_name)
        query = get_sql(self.original_sql, dialect="duckdb")
        uuidstr = _get_temp_table_name()
        full_query = f"""CREATE TEMP VIEW {uuidstr} AS {query};
                         COPY (SELECT *FROM {uuidstr})  
                         TO '{file_name}' (FORMAT JSON, Array True); 
                         DROP VIEW {uuidstr};"""
        await run_in_threadpool(self.con.execute, full_query)


def match_25(
    table: str,
    field: str,
    search_text: str,
    fields: Optional[str] = None,
    alias: Optional[str] = None,
):
    f_args = [ex.column(field), ex.convert(search_text)]
    if fields is not None:
        f_args.append(
            ex.PropertyEQ(
                this=ex.Column(this=ex.Identifier(this="fields", quoted=False)),
                expression=ex.convert(fields),
            )
        )

    fn = ex.Dot(
        this=ex.to_identifier("fts_main_" + table),
        expression=ex.func("match_bm25", *f_args, dialect="duckdb"),
    )
    if alias:
        return fn.as_(alias)
    return fn


class DuckDbExecutionContextBase(ExecutionContext):
    def __init__(
        self,
        con: duckdb.DuckDBPyConnection,
        chunk_size: int,
    ):
        super().__init__(chunk_size=chunk_size, engine_name="duckdb")
        self.con = con.cursor()
        for ins in DUCK_INIT_SCRIPTS:
            self.con.execute(ins)
        self.res_con = None
        self.persistance_file_name = None
        self.array_contains_func = "array_contains"

    @property
    def dialect(self):
        return "duckdb"

    @property
    def supports_view_creation(self) -> bool:
        return True

    def register_arrow(
        self,
        name: str,
        ds: Union[pyarrow.dataset.Dataset, pyarrow.Table],
    ):
        self.con.register(name, ds)

    def close(self):
        self.con.close()

    def execute_sql(
        self,
        sql: Union[
            ex.Query,
            str,
        ],
    ) -> DuckDBResultData:
        if self.persistance_file_name is not None:
            self.res_con = duckdb.connect(
                self.persistance_file_name, read_only=True, config=DUCK_CONFIG
            )

            return DuckDBResultData(
                sql,
                self.res_con,
                self.chunk_size,
            )
        return DuckDBResultData(
            sql,
            con=self.con,
            chunk_size=self.chunk_size,
        )

    def search_score_function(
        self,
        source_view: str,
        search_text: str,
        search_config: SearchConfig,
        alias: Optional[str],
    ):
        fields = ",".join(search_config.columns)
        return match_25(
            source_view,
            "__search_id",
            search_text,
            fields=fields,
            alias=alias,
        )

    def distance_m_function(
        self,
        lat1: ex.Expression,
        lon1: ex.Expression,
        lat2: ex.Expression,
        lon2: ex.Expression,
    ):
        return ex.func("haversine", lat1, lon1, lat2, lon2)

    def json_function(self, term: ex.Expression, assure_string=False):
        fn = ex.func("to_json", term)
        if assure_string:
            return ex.cast(fn, ex.DataType.Type.VARCHAR)
        return fn

    def init_search(
        self,
        source_view: str,
        search_configs: list[SearchConfig],
    ):
        modified_date = self.modified_dates[source_view]
        if modified_date is None:
            return
        persistence_name = source_view
        search_columns = []
        for cfg in search_configs:
            search_columns = search_columns + cfg.columns
        scc = ", ".join([f"{sc}" for sc in search_columns])
        pk_name = "__search_id"
        real_query = f"CREATE TABLE search_con.{persistence_name} AS SELECT ROW_NUMBER() OVER () AS __search_id, * FROM {source_view} s"
        persitance_path = os.getenv("TEMP", "/tmp/")
        persistance_file_name = os.path.join(
            persitance_path, persistence_name + ".duckdb"
        )

        if (
            not os.path.exists(persistance_file_name)
            or datetime.fromtimestamp(
                os.path.getmtime(persistance_file_name), tz=timezone.utc
            )
            < modified_date.astimezone(timezone.utc)
            or datetime.fromtimestamp(
                os.path.getmtime(persistance_file_name), tz=timezone.utc
            )
            < datetime.fromisoformat("2023-05-19").astimezone(
                timezone.utc
            )  # before duckdb upgrade
        ):
            persistance_file_name_temp = persistance_file_name + "_temp"
            if os.path.exists(persistance_file_name_temp):
                os.remove(persistance_file_name_temp)
            search_con = duckdb.connect(
                persistance_file_name_temp, read_only=False, config=DUCK_CONFIG
            )

            search_con.commit()  # create empty duck file
            search_con.close()
            self.con.execute(f"ATTACH '{persistance_file_name_temp}' AS search_con")
            self.con.execute(real_query)
            self.con.execute("DETACH search_con")
            search_con = duckdb.connect(
                persistance_file_name_temp, read_only=False, config=DUCK_CONFIG
            )
            scc = ", ".join([f"'{sc}'" for sc in search_columns])
            search_con.execute(
                f"PRAGMA create_fts_index('{persistence_name}', '{pk_name}', {scc})"
            )
            search_con.close()
            if os.path.exists(persistance_file_name):
                os.remove(persistance_file_name)
            os.rename(persistance_file_name_temp, persistance_file_name)
        if not self.persistance_file_name:
            self.con.load_extension("fts")
        self.persistance_file_name = persistance_file_name

    def init_spatial(self):
        # we don't need spatial extension since haversine is good enough
        self.con.execute(
            """CREATE FUNCTION haversine(lat1, lng1, lat2, lng2) 
    AS ( 6371000 * acos( cos( radians(lat1) ) *
       cos( radians(lat2) ) * cos( radians(lng2) - radians(lng1) ) +
       sin( radians(lat1) ) * sin( radians(lat2) ) ) 
    )"""
        )

    def register_datasource(
        self,
        target_name: str,
        source_table_name: Optional[str],
        uri: SourceUri,
        file_type: FileTypes,
        filters: Optional[FilterType],
        meta_only: bool = False,
    ):
        self.modified_dates[target_name] = self.get_modified_date(uri, file_type)

        remote_uri, remote_opts = uri.get_uri_options(flavor="original")
        duckdb_apply_storage_options(
            self.con,
            remote_uri,
            remote_opts,
            use_fsspec=os.getenv("DUCKDB_DELTA_USE_FSSPEC", "0") == "1",
        )

        if file_type == "json":
            self.con.execute(
                f"""CREATE VIEW {target_name} as 
                    SELECT *FROM read_json_auto('{remote_uri}', format='array')
                 """
            )
            return
        if file_type == "ndjson":
            self.con.execute(
                f"""CREATE VIEW {target_name} as 
                    SELECT *FROM read_json_auto('{remote_uri}', format='newline_delimited')
                """
            )
            return
        if file_type == "parquet":
            self.con.execute(
                f"""CREATE VIEW  {target_name} as 
                    SELECT *FROM read_parquet('{remote_uri}')
                """
            )
            return
        if file_type == "csv":
            self.con.execute(
                f"""CREATE VIEW  {target_name} as 
                            SELECT *FROM read_csv_auto('{remote_uri}', delim=',', header=True)"""
            )
            return
        if file_type == "delta" and uri.exists():
            ab_uri, uri_opts = uri.get_uri_options(flavor="original")
            duckdb_create_view_for_delta(
                self.con,
                ab_uri,
                target_name,
                storage_options=uri_opts,
                conditions=filters,
                use_fsspec=os.getenv("DUCKDB_DELTA_USE_FSSPEC", "0") == "1",
            )
            return
        if file_type == "duckdb":
            self.con.execute(
                f"ATTACH '{remote_uri}' AS  {target_name}_duckdb (READ_ONLY);"
            )
            self.con.execute(
                f"CREATE VIEW  {target_name} as SELECT *FROM  {target_name}_duckdb.{source_table_name or target_name}"
            )
            return
        if file_type == "sqlite":
            self.con.execute(
                f"ATTACH '{remote_uri}' AS  {target_name}_sqlite (TYPE sqlite, READ_ONLY);"
            )
            self.con.execute(
                f"CREATE VIEW  {target_name} as SELECT *FROM  {target_name}_sqlite.{source_table_name or target_name}"
            )
            return

        return super().register_datasource(
            target_name, source_table_name, uri, file_type, filters
        )

    def list_tables(self) -> ResultData:
        return self.execute_sql(
            """SELECT table_name as name, table_type 
               from information_schema.tables where table_schema='main'
            """
        )

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        if self.res_con:
            self.res_con.__exit__(*args, **kwargs)


class DuckDbExecutionContext(DuckDbExecutionContextBase):
    def __init__(self, chunk_size: int):
        super().__init__(duckdb.connect(config=DUCK_CONFIG), chunk_size=chunk_size)

    def __enter__(self):
        super().__enter__()
        self.con.__enter__()
        self.con.execute("SET memory_limit='500MB'")
        self.con.execute(f"SET threads={int(multiprocessing.cpu_count() / 2)}")
        self.con.execute("SET default_null_order='nulls_first'")  # align with polars
        return self

    def __exit__(self, *args, **kwargs):
        super().__exit__(*args, **kwargs)
        self.con.close()

    def close(self):
        self.con.close()



import dataclasses
import bmsdna.lakeapi
import fastapi
import test_server
import os
import asyncio

no_sql = os.getenv("NO_SQL_SERVER", "0") == "1"
sql_server = test_server.start_mssql_server() if not no_sql else None
print("after start mssql")
app = fastapi.FastAPI()


def init():
    def_cfg = bmsdna.lakeapi.get_default_config()  # Get default startup config
    cfg = dataclasses.replace(
        def_cfg, enable_sql_endpoint=True, data_path="tests/data"
    )  # Use dataclasses.replace to set the properties you want
    sti = bmsdna.lakeapi.init_lakeapi(
        app, True, cfg, "config_test.yml"
    )  # Enable it. The first parameter is the FastAPI instance, the 2nd one is the basic config and the third one the config of the tables


init()



import sys
import csv

from fastapi.testclient import TestClient

sys.path.append(".")

engines = ["duckdb"]


def test_duckdb_file_type(client: TestClient):
    for _ in range(2):
        response = client.get("/api/v1/test/fake_duck?limit=1&format=json")
        assert response.status_code == 200


def test_duckdb_file_type_limit_100(client: TestClient):
    for _ in range(2):
        response = client.get("/api/v1/test/fake_duck?limit=100&format=json")
        assert response.status_code == 200
        assert len(response.json()) == 100


def test_duckdb_fruits_fruit_param(client: TestClient):
    for _ in range(2):
        response = client.get(
            "/api/v1/test/fruits_duck?limit=2&format=json&fruits=banana"
        )
        assert response.status_code == 200
        assert response.json() == [
            {"A": 1, "fruits": "banana", "B": 5, "cars": "beetle"},
            {"A": 2, "fruits": "banana", "B": 4, "cars": "audi"},
        ]


def test_duckdb_fruits_car_param(client: TestClient):
    for _ in range(2):
        response = client.get(
            "/api/v1/test/fruits_duck?limit=2&format=json&cars=lamborghini"
        )
        assert response.status_code == 200
        assert response.json() == [
            {"A": 0, "fruits": "apple", "B": 5, "cars": "lamborghini"}
        ]


def test_data_csv(client: TestClient):
    for _ in range(2):
        response = client.get("/api/v1/test/fruits_duck?limit=1&format=csv&cars=audi")
        assert response.status_code == 200
        txt = response.text

        reader = csv.DictReader(txt.splitlines())
        line1 = reader.__next__()
        assert line1 == {"A": "2", "fruits": "banana", "B": "4", "cars": "audi"}



from fastapi.testclient import TestClient


def test_simple_customers(client: TestClient):
    response = client.get("/api/v1/sqlite/sqlite_customers?format=json&limit=50")
    assert response.status_code == 200
    tables = response.json()
    assert len(tables) == 50


def test_filter_country(client: TestClient):
    response = client.get(
        "/api/v1/sqlite/sqlite_customers?format=json&limit=100&Country=Germany"
    )
    assert response.status_code == 200
    tables = response.json()
    assert len(tables) == 4


def test_metadata_detail(client: TestClient):
    response = client.get("/api/v1/sqlite/sqlite_customers/metadata_detail")
    assert response.status_code == 200



from bmsdna.lakeapi.context.df_base import (
    ExecutionContext,
    FileTypeNotSupportedError,
    ResultData,
    get_sql,
)

import pyarrow as pa
from typing import List, Optional, Tuple, Union, cast, Any, TYPE_CHECKING

if TYPE_CHECKING:
    import polars as pl
from bmsdna.lakeapi.core.types import FileTypes, OperatorType
from deltalake2db import FilterType
import pyarrow.dataset
import sqlglot.expressions as ex
from sqlglot import from_, parse_one


from uuid import uuid4
import json
from .source_uri import SourceUri
from sqlglot.dialects.postgres import Postgres


class Polars(Postgres):
    class Generator(Postgres.Generator):
        NULL_ORDERING_SUPPORTED = None


polars_dialect = Polars()

try:
    import polars as pl  # we want to lazy import in case we one day no longer rely on polars if only duckdb is needed

    PL_TO_ARROW = {
        pl.Int8: pa.int8(),
        pl.Int16: pa.int16(),
        pl.Int32: pa.int32(),
        pl.Int64: pa.int64(),
        pl.UInt8: pa.uint8(),
        pl.UInt16: pa.uint16(),
        pl.UInt32: pa.uint32(),
        pl.UInt64: pa.uint64(),
        pl.Float32: pa.float32(),
        pl.Float64: pa.float64(),
        pl.Date: pa.date32(),
        pl.Boolean: pa.bool_(),
        pl.Utf8: pa.large_string(),
        pl.Binary: pa.binary(),
        pl.Categorical: pa.large_string(),
    }
except ImportError:
    PL_TO_ARROW = {}
    pass


class PolarsResultData(ResultData):
    def __init__(
        self,
        sql: Union[ex.Query, str],
        chunk_size: int,
        to_register: dict[str, pl.LazyFrame | pl.DataFrame] = {},
    ):
        super().__init__(chunk_size=chunk_size)
        self.sql = sql
        self._df = None
        self.random_name = "tbl_" + str(uuid4()).replace("-", "")
        self.registred_df = False
        self.to_register = to_register
        self._ctx = None

    def __exit__(self, *args, **kwargs):
        if self._ctx is not None:
            self._ctx.unregister(self._ctx.tables())
        self.to_register = {}
        super().__exit__(*args, **kwargs)

    def get_sql_context(self):
        if self._ctx is not None:
            self._ctx.unregister(self._ctx.tables())
            self._ctx = None
        import polars as pl

        ctx = pl.SQLContext()
        for k, v in self.to_register.items():
            ctx.register(k, v)
        self._ctx = ctx
        return ctx

    def get_df(self):
        if self._df is None:
            real_sql = get_sql(self.sql, dialect=polars_dialect)
            self._df = self.get_sql_context().execute(real_sql)
        return self._df

    async def get_df_collected(self) -> "pl.DataFrame":
        _df = self.get_df()
        if isinstance(_df, pl.LazyFrame):
            _df = await _df.collect_async()
            self._df = _df
        return _df

    def columns(self):
        if self._df is None:
            _df = pl.DataFrame(
                [],
                schema=self.get_sql_context()
                .execute(get_sql(self.sql, limit=0, dialect=polars_dialect))
                .collect_schema(),
            )
            return _df.columns
        return self._df.columns

    def query_builder(self) -> ex.Select:
        if not isinstance(self.sql, str):
            return from_(self.sql.subquery(alias="s1"))
        else:
            return from_(
                cast(ex.Select, parse_one(self.sql, dialect=polars_dialect)).subquery(
                    alias="s1"
                )
            )

    def arrow_schema(self) -> pa.Schema:
        if self._df is None:
            _df = pl.DataFrame(
                [],
                schema=self.get_sql_context()
                .execute(get_sql(self.sql, limit=0, dialect=polars_dialect))
                .collect_schema(),
            )
        else:
            _df = pl.DataFrame([], self._df.collect_schema())
        return _df.to_arrow().schema

    async def to_pandas(self):
        return (await self.get_df_collected()).to_pandas()

    async def to_pylist(self):
        return (await self.get_df_collected()).to_dicts()

    async def to_arrow_recordbatch(self, chunk_size: int = 10000):
        return (
            (await self.get_df_collected())
            .to_arrow()
            .to_reader(max_chunksize=chunk_size)
        )

    async def write_parquet(self, file_name: str):
        (await self.get_df_collected()).write_parquet(file_name, use_pyarrow=True)

    async def write_json(self, file_name: str):
        (await self.get_df_collected()).write_json(file_name)

    async def write_csv(self, file_name: str, *, separator: str):
        (await self.get_df_collected()).write_csv(file_name, separator=separator)

    async def write_nd_json(self, file_name: str):
        (await self.get_df_collected()).write_ndjson(file_name)


class PolarsExecutionContext(ExecutionContext):
    def __init__(self, chunk_size: int):
        super().__init__(chunk_size=chunk_size, engine_name="polars")
        import polars as pl

        self.len_func = "length"
        # self.sql_context = sql_context or pl.SQLContext()
        self._to_register: dict[str, pl.LazyFrame | pl.DataFrame] = {}

    @property
    def dialect(self):
        return polars_dialect

    @property
    def supports_view_creation(self) -> bool:
        return True

    def register_arrow(
        self,
        name: str,
        ds: Union[pyarrow.dataset.Dataset, pyarrow.Table],
    ):
        import polars as pl

        if isinstance(ds, pyarrow.dataset.Dataset):
            self._to_register[name] = pl.scan_pyarrow_dataset(ds)
        else:
            self._to_register[name] = pl.from_arrow(ds)  # type: ignore

    def close(self):
        pass

    def json_function(self, term: ex.Expression, assure_string=False):
        raise NotImplementedError()

    def jsonify_complex(
        self, query: ex.Query, complex_cols: list[str], columns: list[str]
    ):
        import polars as pl

        old_query = query.select(*[ex.column(c, quoted=True) for c in columns])
        if len(complex_cols) == 0:
            return old_query

        rd = PolarsResultData(old_query, self.chunk_size, self._to_register.copy())
        df = rd.get_sql_context().execute(old_query.sql(polars_dialect))

        def to_json(x):
            if isinstance(x, pl.Series):
                return json.dumps(x.to_list())
            return json.dumps(x)

        map_cols = [
            pl.col(c).map_elements(to_json, return_dtype=pl.Utf8).alias(c)
            for c in complex_cols
        ]
        df = df.with_columns(map_cols)
        nt_id = "tmp_" + str(uuid4())
        self._to_register[nt_id] = df
        return from_(ex.to_identifier(nt_id)).select(
            *[ex.column(c, quoted=True) for c in columns]
        )

    def register_datasource(
        self,
        target_name: str,
        source_table_name: Optional[str],
        uri: SourceUri,
        file_type: FileTypes,
        filters: Optional[FilterType],
        meta_only: bool = False,
    ):
        import polars as pl

        ab_uri, uri_opts = uri.get_uri_options(flavor="object_store")

        self.modified_dates[target_name] = self.get_modified_date(uri, file_type)
        match file_type:
            case "delta":
                from deltalake2db.protocol_check import DeltaProtocolError

                try:
                    db_uri, db_opts = uri.get_uri_options(flavor="original")
                    from deltalake2db import polars_scan_delta, get_polars_schema

                    if meta_only:
                        schema = get_polars_schema(db_uri, storage_options=db_opts)
                        df = pl.DataFrame(schema=schema)
                    else:
                        df = polars_scan_delta(
                            db_uri,
                            storage_options=db_opts,
                            conditions=filters,
                        )
                except DeltaProtocolError as de:
                    raise FileTypeNotSupportedError(
                        f"Delta table version {ab_uri} not supported"
                    ) from de
            case "parquet":
                df = pl.scan_parquet(ab_uri, storage_options=uri_opts)
            case "arrow":
                df = pl.scan_ipc(ab_uri, storage_options=uri_opts)
            case "csv" if uri_opts is None:
                df = pl.scan_csv(ab_uri)
            case "csv" if uri_opts is not None:
                df = pl.read_csv(ab_uri)
            case "json" if uri_opts is None:
                df = cast(pl.LazyFrame, pl.read_json(ab_uri))
            case "ndjson" if uri_opts is None:
                df = pl.scan_ndjson(ab_uri)
            case "sqlite" if uri_opts is None:
                query = "SELECT * FROM " + (source_table_name or target_name)

                df = pl.read_database_uri(
                    query=query, uri="sqlite://" + ab_uri, engine="adbc"
                )
            case "duckdb" if uri_opts is None:
                query = "SELECT * FROM " + (source_table_name or target_name)
                import duckdb

                with duckdb.connect(ab_uri, read_only=True) as con:
                    t = con.execute(query).fetch_arrow_table()

                    df = cast(pl.LazyFrame, pl.from_arrow(t))
            # case "odbc": to be tested, attention on security!
            #             #    query = "SELECT * FROM " + (source_table_name or target_name)
            #    df = pl.read_database(query=query, connection=uri.uri)
            case _:
                raise FileTypeNotSupportedError(f"Not supported file type {file_type}")

        self._to_register[target_name] = df

    def execute_sql(
        self,
        sql: Union[
            ex.Query,
            str,
        ],
    ) -> PolarsResultData:
        return PolarsResultData(sql, self.chunk_size, self._to_register.copy())

    def list_tables(self) -> ResultData:
        return self.execute_sql("SHOW TABLES")

    def __enter__(self):
        return self

    def __exit__(self, *args):
        pass



from dataclasses import dataclass
from typing import Any, List, Literal, cast, Sequence
from datetime import datetime, date, time, timedelta
from decimal import Decimal
from typing import Type, Optional
from pydantic import BaseModel


Engines = Literal["duckdb", "polars", "odbc"]


FileTypes = Literal[
    "delta",
    "parquet",
    "arrow",
    "arrow-stream",
    "csv",
    "json",
    "ndjson",
    "odbc",
    "sqlite",
    "duckdb",
]
OutputFileType = Literal[
    "json",
    "ndjson",
    "parquet",
    "csv",
    "csv4excel",
    "xlsx",
    "feather",
    "html",
    "scsv",
    "xml",
    "ipc",
    "arrow",
    "arrow-stream",
]
OperatorType = Literal[
    "<",
    "=",
    ">",
    ">=",
    "<=",
    "<>",
    "contains",
    "in",
    "not contains",
    "not in",
    "not null",
    "null",
    "between",
    "not between",
    "has",
    "startswith",
]
DeltaOperatorTypes = Literal["<", "=", ">", ">=", "<=", "in", "not in"]
PolaryTypeFunction = Literal[
    "sum",
    "count",
    "mean",
    "median",
    "min",
    "max",
    "first",
    "std",
    "n_unique",
    "distinct",
    "mode",
    "null_count",
]

CONFIG_DTYPE_MAP: dict[str, Type] = {
    "struct": dict,
    "string": str,
    "integer": int,
    "short": int,
    "byte": int,
    "float": float,
    "array": list,
    "double": Decimal,
    "timestamp": datetime,
    "date": date,
    "time": time,
    "duration": timedelta,
    "str": str,
    "int": int,
    "bool": bool,
    "boolean": bool,
}


DTYPE_MAP = {
    "struct": dict,
    "string": str,
    "integer": int,
    "float": float,
    "array": list,
    "double": Decimal,
    "timesamp": datetime,
    "date": date,
    "time": time,
}


class MetadataSchemaFieldType(BaseModel):
    type_str: str
    orig_type_str: str
    fields: "Optional[list[MetadataSchemaField]]"
    inner: "Optional[MetadataSchemaFieldType]"


class MetadataSchemaField(BaseModel):
    name: str
    type: MetadataSchemaFieldType
    max_str_length: Optional[int] = None


@dataclass
class SearchConfig:
    name: str
    columns: List[str]


@dataclass
class NearbyConfig:
    name: str
    lat_col: str
    lon_col: str


@dataclass
class Param:
    name: str
    combi: Optional[Optional[List[str]]] = None
    default: Optional[str] = None
    required: Optional[bool] = False
    operators: Optional[list[OperatorType]] = None
    colname: Optional[str] = None

    @property
    def real_default(self) -> str:
        import ast

        self._real_default = ast.literal_eval(self.default) if self.default else None
        return cast(str, self._real_default)


class MetadataDetailResult(BaseModel):
    partition_values: Optional[Sequence[dict[str, Any]]] = None
    partition_columns: List[str]
    max_string_lengths: dict[str, Optional[int]]
    data_schema: list[MetadataSchemaField]
    delta_meta: Optional[dict] = None
    delta_schema: Any = None
    parameters: Optional[List[Any]] = None
    search: Optional[List[Any]] = None
    nearby: Optional[List[Any]] = None
    modified_date: datetime | None = None


MetadataSchemaFieldType.model_rebuild()
MetadataSchemaField.model_rebuild()
MetadataDetailResult.model_rebuild()



import duckdb
import requests

with open("azure.duckdb_extension.gz", "wb") as f:
    f.write(
        requests.get(
            "http://extensions.duckdb.org/v0.10.0/windows_amd64/azure.duckdb_extension.gz"
        ).content
    )

emulator_con_str = "DefaultEndpointsProtocol=http;AccountName=devstoreaccount1;AccountKey=Eby8vdM02xNOcqFlqUwJPLlmEtlCDXJ1OUzFT50uSRZ6IFsuFq2UVErCz4I6tq/K1SZFPTOtr/KBHBeksoGMGw==;BlobEndpoint=http://127.0.0.1:10000/devstoreaccount1;QueueEndpoint=http://127.0.0.1:10001/devstoreaccount1;TableEndpoint=http://127.0.0.1:10002/devstoreaccount1;"

with duckdb.connect() as con:
    con.execute(
        f"""FORCE INSTALL azure;
        INSTALL azure;
LOAD azure;
CREATE SECRET secret2 (
    TYPE AZURE,
    CONNECTION_STRING  '{emulator_con_str}'
);"""
    )
    with con.cursor() as cur:
        cur.execute(
            "SELECT count(*) FROM read_parquet('azure://testlake/td/faker.parquet')"
        )
        print(cur.fetchall())



import hashlib
import mimetypes
import os
import tempfile
from enum import Enum
from typing import Union
from uuid import uuid4
import pyarrow as pa
from starlette.background import BackgroundTask
from starlette.datastructures import URL, QueryParams
from starlette.responses import Response, StreamingResponse
from email.utils import formatdate
from bmsdna.lakeapi.context.df_base import ExecutionContext, ResultData
from bmsdna.lakeapi.core.config import BasicConfig
from bmsdna.lakeapi.core.log import get_logger
from bmsdna.lakeapi.core.types import OutputFileType
import typing
from urllib.parse import quote
from mimetypes import guess_type
from starlette.concurrency import iterate_in_threadpool
import sqlglot.expressions as ex
import anyio

from bmsdna.lakeapi.utils.async_utils import _async

logger = get_logger(__name__)


class OutputFormats(Enum):
    CSV = 2
    CSV4EXCEL = 12
    SEMI_CSV = 11
    XLSX = 3
    HTML = 4
    ARROW_IPC = 5
    ND_JSON = 6
    PARQUET = 7
    JSON = 8
    XML = 9
    ORC = 10
    ARROW_STREAM = 14


async def parse_format(accept: Union[str, OutputFileType]) -> tuple[OutputFormats, str]:
    realaccept = accept.split(";")[0].strip().lower()
    if realaccept == "text/csv" or realaccept == "csv":
        return (OutputFormats.CSV, ".csv")
    elif realaccept == "text/csv+semicolon" or realaccept == "scsv":
        return (OutputFormats.SEMI_CSV, ".csv")
    elif realaccept == "x.text/csv+excel" or realaccept == "csv4excel":
        return (OutputFormats.CSV4EXCEL, ".csv")
    elif realaccept == "application/xml" or realaccept == "xml":
        return (OutputFormats.XML, ".xml")
    elif realaccept == "orc":
        return (OutputFormats.ORC, ".orc")
    elif realaccept == mimetypes.guess_type("file.xlsx") or realaccept == "xlsx":
        return (OutputFormats.XLSX, ".xlsx")
    elif realaccept == "text/html" or realaccept == "html":
        return (OutputFormats.HTML, ".html")
    elif (
        realaccept == "application/vnd.apache.arrow.stream"
        or realaccept == "arrow-stream"
    ):
        return (OutputFormats.ARROW_STREAM, "")

    elif (
        realaccept == "application/x-arrow"
        or realaccept == "application/x-arrow"
        or realaccept == "application/vnd.apache.arrow.file"
        or realaccept == "arrow"
        or realaccept == "feather"
        or realaccept == "ipc"
    ):
        return (OutputFormats.ARROW_IPC, ".arrow")
    elif (
        realaccept == "application/json+newline"
        or realaccept == "application/jsonl"
        or realaccept == "ndjson"
    ):
        return (OutputFormats.ND_JSON, ".ndjson")
    elif realaccept == "application/parquet" or realaccept == "parquet":
        return (OutputFormats.PARQUET, ".parquet")
    else:
        return (OutputFormats.JSON, ".json")


async def _write_frame(
    url: URL,
    content: ResultData,
    format: OutputFormats,
    out: str,
    basic_config: BasicConfig,
    csv_separator: str | None = None,
    charset: str = "utf-8",
) -> list[str]:
    if csv_separator == "\\t":
        csv_separator = "\t"
    is_utf_8 = charset.lower() == "utf-8"
    if format == OutputFormats.CSV:
        await _async(
            content.write_csv(
                out + ("" if is_utf_8 else "_u8"), separator=csv_separator or ","
            )
        )
    elif format == OutputFormats.SEMI_CSV:
        await _async(
            content.write_csv(
                out + ("" if is_utf_8 else "_u8"), separator=csv_separator or ";"
            )
        )
    elif format == OutputFormats.CSV4EXCEL:  # need to write sep=, on first line
        await _async(
            content.write_csv(
                out + ("" if is_utf_8 else "_u8"), separator=csv_separator or ","
            )
        )

    elif format == OutputFormats.XLSX:
        import polars as pl

        ds = pl.from_arrow(await _async(content.to_arrow_recordbatch()))
        assert isinstance(ds, pl.DataFrame)
        ds.write_excel(out, autofit=True)
        is_utf_8 = True
    elif format == OutputFormats.HTML:
        (await _async(content.to_pandas())).to_html(
            out + ("" if is_utf_8 else "_u8"), index=False
        )

    elif format == OutputFormats.XML:
        (await _async(content.to_pandas())).to_xml(
            out + ("" if is_utf_8 else "_u8"), index=False, parser="etree"
        )

    elif format == OutputFormats.ARROW_IPC:
        is_utf_8 = True
        with await _async(content.to_arrow_recordbatch(content.chunk_size)) as batches:
            with pa.OSFile(out, "wb") as sink:
                with pa.ipc.new_file(sink, batches.schema) as writer:
                    for batch in batches:
                        writer.write(batch)

    elif format == OutputFormats.ARROW_STREAM:
        is_utf_8 = True
        with await _async(content.to_arrow_recordbatch(content.chunk_size)) as batches:
            with pa.OSFile(out, "wb") as sink:
                with pa.ipc.new_stream(sink, batches.schema) as writer:
                    for batch in batches:
                        writer.write_batch(batch)

    elif format == OutputFormats.ND_JSON:
        await _async(content.write_nd_json(out + ("" if is_utf_8 else "_u8")))
    elif format == OutputFormats.PARQUET:
        await _async(content.write_parquet(out))
        is_utf_8 = True
    else:
        await content.write_json(out + ("" if is_utf_8 else "_u8"))
    if not is_utf_8:
        with (
            open(out, mode="wb") as f
        ):  # excel wants utf-16le which polars does not support. therefore we need reencoding
            if format == OutputFormats.CSV4EXCEL:
                f.write(b"sep=,\n")  # add utf-8 bom at beginning
            with open(out + "_u8", mode="r", encoding="utf-8") as c8:
                line = c8.readline()
                while line != "":
                    f.write(line.encode(charset))
                    line = c8.readline()
            return [out + "_u8"]
    return []


Content = typing.Union[str, bytes]
SyncContentStream = typing.Iterator[Content]
AsyncContentStream = typing.AsyncIterable[Content]
ContentStream = typing.Union[AsyncContentStream, SyncContentStream]


class StreamingResponseWCharset(StreamingResponse):
    """StreamingResponseWCharset
    Combines file response with stream response
    """

    body_iterator: AsyncContentStream

    def __init__(
        self,
        content: ContentStream,
        status_code: int = 200,
        headers: typing.Optional[typing.Mapping[str, str]] = None,
        media_type: typing.Optional[str] = None,
        background: typing.Optional[BackgroundTask] = None,
        filename: typing.Optional[str] = None,
        stat_result: typing.Optional[os.stat_result] = None,
        method: typing.Optional[str] = None,
        content_disposition_type: str = "attachment",
        *args,
        **kwargs,
    ):
        if isinstance(content, typing.AsyncIterable):
            self.body_iterator = content  # type: ignore
        else:
            self.body_iterator = iterate_in_threadpool(content)  # type: ignore

        # taking over from FileResponse
        self.status_code = status_code
        self.filename = filename
        self.send_header_only = method is not None and method.upper() == "HEAD"
        if "charset" in kwargs:
            self.charset = kwargs.pop("charset")
        if media_type is None:
            media_type = guess_type(filename or "text/pain")[0] or "text/plain"

        self.media_type = media_type + (
            "; charset=" + self.charset
            if self.charset and "charset" not in media_type
            else ""
        )
        self.background = background
        self.init_headers(headers)
        if self.filename is not None:
            content_disposition_filename = quote(self.filename)
            if content_disposition_filename != self.filename:
                content_disposition = "{}; filename*=utf-8''{}".format(
                    content_disposition_type, content_disposition_filename
                )
            else:
                content_disposition = '{}; filename="{}"'.format(
                    content_disposition_type, self.filename
                )
            self.headers.setdefault("content-disposition", content_disposition)
        self.stat_result = stat_result
        if stat_result is not None:
            self.set_stat_headers(stat_result)

    def set_stat_headers(self, stat_result: os.stat_result) -> None:
        content_length = str(stat_result.st_size)
        last_modified = formatdate(stat_result.st_mtime, usegmt=True)
        etag_base = str(stat_result.st_mtime) + "-" + str(stat_result.st_size)
        etag = hashlib.md5(etag_base.encode()).hexdigest()

        self.headers.setdefault("content-length", content_length)
        self.headers.setdefault("last-modified", last_modified)
        self.headers.setdefault("etag", etag)


class TempFileWrapper:  # does not open the file which is important on windows
    def __init__(self, path: str) -> None:
        self.path = path

    @property
    def name(self):
        return self.path

    def close(self):
        os.unlink(self.path)


def get_temp_file(extension: str):
    if os.name == "nt":
        temp_file = TempFileWrapper(os.environ["TEMP"] + "/" + str(uuid4()))
    else:
        temp_file = tempfile.NamedTemporaryFile(delete=True, suffix=extension)
    return temp_file


async def create_response(
    url: URL,
    query_params: QueryParams,
    accept: str,
    context: ExecutionContext,
    sql: ex.Query | str,
    basic_config: BasicConfig,
    charset: str | None = None,
    close_context=False,
):
    headers = {}
    format, extension = await parse_format(accept)

    charset = charset or ("utf-16-le" if format == OutputFormats.CSV4EXCEL else "utf-8")

    content_dispositiont_type = "attachment"
    filename = "file" + extension
    media_type = (
        "text/csv"
        if extension == ".csv"
        else mimetypes.guess_type("file" + extension)[0]
    )

    if format == OutputFormats.JSON:
        with context.execute_sql(sql) as res:
            return Response(
                content=await _async(res.to_json()),
                headers=headers,
                media_type=(media_type or "application/json") + "; charset=" + charset,
            )
    if format == OutputFormats.ND_JSON:
        with context.execute_sql(sql) as res:
            return Response(
                content=await _async(res.to_ndjson()),
                headers=headers,
                media_type="application/json-nd; charset=" + charset,
            )

    if format in [
        OutputFormats.JSON,
        OutputFormats.ND_JSON,
        OutputFormats.CSV,
        OutputFormats.SEMI_CSV,
        OutputFormats.CSV4EXCEL,
    ]:
        content_dispositiont_type = "inline"
        filename = None

    temp_file = get_temp_file(extension)

    async def response_stream(context: ExecutionContext, sql, url, format):
        chunk_size = 64 * 1024
        with context.execute_sql(sql) as content:
            additional_files = await _write_frame(
                url,
                content,
                format,
                temp_file.name,
                basic_config,
                query_params.get("$csv_separator", None),
                charset,
            )

        async with await anyio.open_file(temp_file.name, mode="rb") as file:
            more_body = True
            while more_body:
                chunk = await file.read(chunk_size)
                more_body = len(chunk) == chunk_size
                yield chunk

    def clean_up():
        if close_context:
            context.close()
        try:
            temp_file.close()
        except FileNotFoundError:
            pass

    return StreamingResponseWCharset(
        content=response_stream(context, sql, url, format),
        headers=headers,
        media_type=media_type,
        content_disposition_type=content_dispositiont_type,
        filename=filename,
        background=BackgroundTask(clean_up),
        charset=charset,
    )



from fastapi.testclient import TestClient
import sys
import pytest

sys.path.append(".")


engines = ("duckdb", "polars")


@pytest.mark.parametrize("e", engines)
def test_no_filter(client: TestClient, e):
    response = client.get(
        f"/api/v1/deltatest/table_w_col_map?limit=50&format=json&%24engine={e}"
    )
    assert response.status_code == 200
    jsd = response.json()
    assert len(jsd) == 50


@pytest.mark.parametrize("e", engines)
def test_col_map_filter(client: TestClient, e):
    response = client.get(
        f"/api/v1/deltatest/table_w_col_map?limit=50&format=json&Super_Name_=John Duncan&%24engine={e}"
    )
    assert response.status_code == 200
    jsd = response.json()
    assert len(jsd) == 1
    for item in jsd:
        assert item["Super Name_"] == "John Duncan"



from typing import Callable
from bmsdna.lakeapi.context.df_base import ExecutionContext
from bmsdna.lakeapi.core.types import Engines


def _duckdb(chunk_size: int):
    from bmsdna.lakeapi.context.df_duckdb import DuckDbExecutionContext

    return DuckDbExecutionContext(chunk_size=chunk_size)


def _polars(chunk_size: int):
    from bmsdna.lakeapi.context.df_polars import PolarsExecutionContext

    return PolarsExecutionContext(chunk_size=chunk_size)


def _odbc(chunk_size: int):
    from bmsdna.lakeapi.context.df_odbc import ODBCExecutionContext

    return ODBCExecutionContext(chunk_size=chunk_size)


engine_registry = {"duckdb": _duckdb, "polars": _polars, "odbc": _odbc}


def register_engine(engine: Engines, factory: Callable[[int], ExecutionContext]):
    engine_registry[engine] = factory


def get_context_by_engine(
    engine: Engines,
    chunk_size: int,
) -> ExecutionContext:
    return engine_registry[engine](chunk_size)


class ExecutionContextManager:
    default_engine: Engines

    def __init__(
        self,
        default_engine: Engines,
        default_chunk_size: int,
    ):
        self.default_engine = default_engine
        self.contexts: dict[str, ExecutionContext] = dict()
        self.default_chunk_size = default_chunk_size

    def get_context(
        self,
        engine: Engines | None,
        chunk_size: int | None = None,
    ):
        real_engine: Engines = engine or self.default_engine
        if real_engine not in self.contexts:
            self.contexts[real_engine] = get_context_by_engine(
                real_engine, chunk_size or self.default_chunk_size
            )
        return self.contexts[real_engine]

    def __enter__(self, *args, **kwargs):
        return self

    def __exit__(self, *args, **kwargs):
        for _, ctx in self.contexts.items():
            ctx.__exit__(*args, **kwargs)
        self.contexts = dict()



import asyncio
import logging
from asyncio import ensure_future
from functools import wraps
from traceback import format_exception
from typing import Any, Callable, Coroutine, Union

from starlette.concurrency import run_in_threadpool

NoArgsNoReturnFuncT = Callable[[], None]
NoArgsNoReturnAsyncFuncT = Callable[[], Coroutine[Any, Any, None]]
NoArgsNoReturnDecorator = Callable[
    [Union[NoArgsNoReturnFuncT, NoArgsNoReturnAsyncFuncT]], NoArgsNoReturnAsyncFuncT
]


def _repeat_every(
    *,
    seconds: float,
    wait_first: bool = False,
    logger: logging.Logger | None = None,
    raise_exceptions: bool = False,
    max_repetitions: int | None = None,
) -> NoArgsNoReturnDecorator:
    """
    This function returns a decorator that modifies a function so it is periodically re-executed after its first call.

    The function it decorates should accept no arguments and return nothing. If necessary, this can be accomplished
    by using `functools.partial` or otherwise wrapping the target function prior to decoration.

    Parameters
    ----------
    seconds: float
        The number of seconds to wait between repeated calls
    wait_first: bool (default False)
        If True, the function will wait for a single period before the first call
    logger: Optional[logging.Logger] (default None)
        The logger to use to log any exceptions raised by calls to the decorated function.
        If not provided, exceptions will not be logged by this function (though they may be handled by the event loop).
    raise_exceptions: bool (default False)
        If True, errors raised by the decorated function will be raised to the event loop's exception handler.
        Note that if an error is raised, the repeated execution will stop.
        Otherwise, exceptions are just logged and the execution continues to repeat.
        See https://docs.python.org/3/library/asyncio-eventloop.html#asyncio.loop.set_exception_handler for more info.
    max_repetitions: Optional[int] (default None)
        The maximum number of times to call the repeated function. If `None`, the function is repeated forever.
    """

    def decorator(
        func: NoArgsNoReturnAsyncFuncT | NoArgsNoReturnFuncT,
    ) -> NoArgsNoReturnAsyncFuncT:
        """
        Converts the decorated function into a repeated, periodically-called version of itself.
        """
        is_coroutine = asyncio.iscoroutinefunction(func)

        @wraps(func)
        async def wrapped() -> None:
            repetitions = 0

            async def loop() -> None:
                nonlocal repetitions
                if wait_first:
                    await asyncio.sleep(seconds)
                while max_repetitions is None or repetitions < max_repetitions:
                    try:
                        if is_coroutine:
                            await func()  # type: ignore
                        else:
                            await run_in_threadpool(func)
                        repetitions += 1
                    except Exception as exc:
                        if logger is not None:
                            formatted_exception = "".join(
                                format_exception(type(exc), exc, exc.__traceback__)
                            )
                            logger.error(formatted_exception)
                        if raise_exceptions:
                            raise exc
                    await asyncio.sleep(seconds)

            ensure_future(loop())

        return wrapped

    return decorator



import sys
import os
import pathlib
from typing import Any, Literal, Optional
import polars as pl
from deltalake import write_deltalake, DeltaTable
import shutil
import pandas as pd
from hashlib import md5
import pyarrow as pa
import json
import duckdb
import sqlite3
from datetime import date
from .utils import create_rows_faker


dir_path = pathlib.Path(os.path.dirname(os.path.realpath(__file__))).parent
sys.path.append(os.path.realpath(f"{dir_path}"))


def delete_folder(path):
    if not os.path.exists(path):
        return  # nothing todo
    if os.path.isfile(path):
        os.remove(path)
    else:
        shutil.rmtree(path)


def store_df_as_delta(
    data: dict[str, list[Any]] | pd.DataFrame | pa.Table,
    data_path: str,
    partition_by: Optional[list[str]] = None,
    *,
    table_properties: Optional[dict[str, str | dict | list]] = None,
    compression: Literal[
        "UNCOMPRESSED", "SNAPPY", "GZIP", "BROTLI", "LZ4", "ZSTD", "LZ4_RAW"
    ]
    | None = None,
) -> pd.DataFrame:
    from deltalake import WriterProperties

    dfp: pa.Table | pd.DataFrame = (
        data
        if isinstance(data, pd.DataFrame) or isinstance(data, pa.Table)
        else pl.DataFrame(data).to_pandas()
    )
    delta_path = "tests/data/" + data_path
    delete_folder(delta_path)

    def _str_or_json(v: str | dict | list):
        if isinstance(v, str):
            return v
        return json.dumps(v)

    write_deltalake(
        delta_path,
        dfp,
        mode="overwrite",
        partition_by=partition_by,
        writer_properties=WriterProperties(compression=compression or "SNAPPY"),
    )
    if table_properties is not None:
        DeltaTable(delta_path).alter.set_table_properties(
            {k: _str_or_json(v) for k, v in table_properties.items()},
            raise_if_not_exists=False,
        )
    assert not isinstance(dfp, dict)
    return dfp if isinstance(dfp, pd.DataFrame) else dfp.to_pandas()


if __name__ == "__main__":
    if not os.path.exists("tests/data/parquet/search.parquet"):
        os.makedirs("tests/data/parquet", exist_ok=True)
        pl.DataFrame(create_rows_faker(1000)).write_parquet(
            "tests/data/parquet/search.parquet"
        )

    df_fruits = store_df_as_delta(
        {
            "A": [1, 2, 3, 4, 5, 0, 9],
            "fruits": [
                "banana",
                "banana",
                "apple",
                "apple",
                "banana",
                "apple",
                "ananas",
            ],
            "B": [5, 4, 3, 2, 1, 5, 9],
            "cars": [
                "beetle",
                "audi",
                "beetle",
                "beetle",
                "beetle",
                "lamborghini",
                "fiat",
            ],
        },
        "delta/fruits",
    )
    store_df_as_delta(df_fruits, "startest/fruits")
    store_df_as_delta(
        {
            "date_field": [
                date(2022, 12, 2),
                date(2022, 12, 5),
                date(2022, 12, 7),
                date(2022, 11, 2),
                date(2022, 2, 2),
                date(2021, 12, 2),
                date(2023, 10, 2),
            ],
            "fruits": [
                "banana",
                "banana",
                "apple",
                "apple",
                "banana",
                "apple",
                "ananas",
            ],
            "B": [5, 4, 3, 2, 1, 5, 9],
            "cars": [
                "beetle",
                "audi",
                "beetle",
                "beetle",
                "beetle",
                "lamborghini",
                "fiat",
            ],
        },
        "delta/dates",
    )

    store_df_as_delta(
        {
            "A": [1, 2, 3, 4, 5, 9],
            "fruits": ["banana", "banana", "apple", "apple", "banana", "ananas"],
            "B": [5, 4, 3, 2, 1, 9],
            "cars": ["beetle", "audi", "beetle", "beetle", "beetle", "fiat"],
            "person": [
                {"name": "tom", "age": 3},
                {"name": "bob", "age": 5},
                {"name": "tim", "age": 7},
                {"name": "john", "age": 1},
                {"name": "marc", "age": 3},
                {"name": "peter", "age": 32},
            ],
            "vitamines": [
                ["A", "B12"],
                [],
                ["C", "B12"],
                ["D", "B12", "C"],
                ["C"],
                ["E", "B12"],
            ],
        },
        "delta/struct_fruits",
        table_properties={
            "lakeapi.config": {
                "params": [
                    {
                        "name": "fruits",
                        "operators": [
                            "not in",
                            "in",
                            "contains",
                            "startswith",
                            "not contains",
                            "<>",
                        ],
                    },
                    {
                        "name": "cars",
                        "operators": [
                            "not in",
                            "in",
                            "contains",
                            "startswith",
                            "not contains",
                            "<>",
                        ],
                    },
                    {
                        "name": "B",
                        "operators": [
                            ">",
                            "<",
                            "<=",
                            ">=",
                            "between",
                            "startswith",
                            "not between",
                        ],
                    },
                ]
            }
        },
    )

    os.makedirs("tests/data/duckdb", exist_ok=True)
    if os.path.exists("tests/data/duckdb/fruits.db"):
        os.remove("tests/data/duckdb/fruits.db")
    con = duckdb.connect("tests/data/duckdb/fruits.db")
    con.execute("DROP TABLE IF EXISTS fruits;")
    con.execute("CREATE TABLE fruits as SELECT * FROM df_fruits;")

    os.makedirs("tests/data/sqlite", exist_ok=True)
    conn = sqlite3.connect("tests/data/sqlite/fruits.sqlite")
    conn.execute("DROP TABLE IF EXISTS fruits;")
    conn.execute("CREATE TABLE fruits (A int, fruits text, B int, cars text);")
    df_fruits.to_sql("fruits", conn, if_exists="replace", index=False)
    conn.commit()
    conn.close()

    fruits_partition_int = df_fruits.copy()
    store_df_as_delta(
        fruits_partition_int,
        "delta/fruits_partition_int",
        partition_by=["A"],
    )

    fruits_partition = df_fruits.copy()
    fruits_partition["my_empty_col"] = pd.Series(
        data=[None for _ in range(0, fruits_partition.shape[0])], dtype="string"
    )
    fruits_partition["fruits_partition"] = fruits_partition["fruits"]
    fruits_partition["cars_md5_prefix_2"] = [
        md5(val.encode("UTF-8")).hexdigest()[:2] for val in fruits_partition["cars"]
    ]
    store_df_as_delta(
        fruits_partition,
        "delta/fruits_partition",
        partition_by=["cars_md5_prefix_2", "cars"],
    )
    store_df_as_delta(
        fruits_partition,
        "startest/fruits_partition",
        partition_by=["cars_md5_prefix_2", "cars"],
    )
    fruits_partition_mod = df_fruits.copy()
    fruits_partition_mod["cars_md5_mod_27"] = [
        str(int(md5(val.encode("UTF-8")).hexdigest(), 16) % 27)
        for val in fruits_partition_mod["cars"]
    ]
    store_df_as_delta(
        fruits_partition_mod,
        "delta/fruits_partition_mod",
        partition_by=["cars_md5_mod_27"],
    )

    df_fruits_nested = pl.from_pandas(df_fruits)
    df_fruits_nested = df_fruits_nested.with_columns(
        pl.struct([df_fruits_nested["fruits"], df_fruits_nested["cars"]]).alias(
            "nested"
        )
    )

    df_fruits_nested = df_fruits_nested.to_pandas()
    store_df_as_delta(df_fruits_nested, "delta/fruits_nested")
    store_df_as_delta(df_fruits_nested, "startest/fruits_nested")
    weather = pl.DataFrame(
        {
            "station": ["Station " + str(x) for x in range(1, 6)],
            "temperatures": [
                "20 5 5 E1 7 13 19 9 6 20",
                "18 8 16 11 23 E2 8 E2 E2 E2 90 70 40",
                "19 24 E9 16 6 12 10 22",
                "E2 E0 15 7 8 10 E1 24 17 13 6",
                "14 8 E0 16 22 24 E1",
            ],
        }
    )
    out = weather.with_columns(
        pl.col("temperatures").str.split(" ")
    )  # thanks for the sample, polars
    store_df_as_delta(out.to_pandas(), "delta/weather", compression="ZSTD")

    csv_path = "tests/data/csv/fruits.csv"
    delete_folder(csv_path)
    os.makedirs(pathlib.Path(csv_path).parent, exist_ok=True)
    df_fruits.to_csv(csv_path)
    df_fruits.to_csv("tests/data/startest/fruits_csv.csv")
    json_path = "tests/data/json/fruits.json"
    delete_folder(json_path)
    os.makedirs(pathlib.Path(json_path).parent, exist_ok=True)
    df_fruits.to_json(json_path, orient="records")

    json_path = "tests/data/ndjson/fruits.ndjson"
    delete_folder(json_path)
    os.makedirs(pathlib.Path(json_path).parent, exist_ok=True)
    df_fruits.to_json(json_path, orient="records", lines=True)

    df_faker = pl.DataFrame(create_rows_faker(100_011)).to_pandas()

    df_faker["name_md5_prefix_2"] = [
        md5(val.encode("UTF-8")).hexdigest()[:1] for val in df_faker["name"]
    ]

    df_faker["name1"] = df_faker["name"]

    print(df_faker)

    os.makedirs("tests/data/duckdb", exist_ok=True)
    if os.path.exists("tests/data/duckdb/faker.db"):
        os.remove("tests/data/duckdb/faker.db")
    con = duckdb.connect("tests/data/duckdb/faker.db")
    con.execute("DROP TABLE IF EXISTS fake;")
    con.execute("CREATE TABLE fake as SELECT * FROM df_faker;")

    for _ in range(10):
        con.execute("insert into fake select * from df_faker;")

    store_df_as_delta(df_faker, "delta/fake", partition_by=None)
    # store_df_as_delta(df_faker, "delta/fake_partition", partition_by=None)

    parquet_path = "tests/data/parquet/fake.parquet"
    delete_folder(parquet_path)
    os.makedirs(pathlib.Path(parquet_path).parent, exist_ok=True)
    df_faker.to_parquet(parquet_path)
    df_faker.to_parquet("tests/data/startest/faker.parquet")
    arrow_path = "tests/data/arrow/fake.arrow"
    delete_folder(arrow_path)
    os.makedirs(pathlib.Path(arrow_path).parent, exist_ok=True)
    df_faker.to_feather(arrow_path)

    df_ns = df_faker.copy()
    from faker import Faker

    fakeit = Faker()
    df_ns["ts"] = pl.Series(
        "ts",
        [fakeit.date_time() for _ in range(0, df_ns.shape[0])],
        pl.Datetime(time_unit="ns"),
    )
    df_ns.to_parquet("tests/data/parquet/fake_ns.parquet")

    import test_server

    if os.getenv("NO_AZURITE_DOCKER", "0") == "0":
        print("start docker azurite")
        test_server.start_azurite()
    else:
        test_server.upload_to_azurite()



import hashlib
import os
from datetime import datetime, date
from typing import (
    Any,
    Iterable,
    List,
    Literal,
    Optional,
    Sequence,
    Tuple,
    Union,
    cast,
    get_args,
    TypeAlias,
)
from bmsdna.lakeapi.context.source_uri import SourceUri
import pyarrow as pa
import pyarrow.parquet
import sqlglot.expressions as ex
from sqlglot import select
from bmsdna.lakeapi.context.df_base import ExecutionContext, ResultData
from bmsdna.lakeapi.core.config import (
    BasicConfig,
    DatasourceConfig,
    Param,
    SelectColumn,
)
from bmsdna.lakeapi.core.log import get_logger
from bmsdna.lakeapi.core.model import get_param_def
from bmsdna.lakeapi.core.types import DeltaOperatorTypes, OperatorType
from deltalake2db.delta_meta_retrieval import (
    DataType,
    field_to_type,
    PrimitiveType,
    get_meta,
    PolarsEngine,
    MetaState as DeltaMetadata,
)
from bmsdna.lakeapi.utils.async_utils import _async
from deltalake2db.filter_by_meta import FilterType, Operator as SupportedOperator

logger = get_logger(__name__)

endpoints = Literal["query", "meta", "request", "sql"]


def select_df(
    df: ex.Query, select_cols: Optional[List[SelectColumn]], exclude_cols: list[str]
) -> ex.Query:
    if select_cols:
        select = [
            ex.column(c.name, quoted=True).as_(c.alias or c.name, quoted=True)
            for c in select_cols
            if c not in (exclude_cols or [])
        ]
        return df.select(*select, append=False)
    else:
        return df.select(ex.Star(), append=False)


def to_pyarrow_schema(schema: DataType) -> pa.DataType:
    fields = []
    if isinstance(schema, str):
        type_map: dict[PrimitiveType, pa.DataType] = {
            "integer": pa.int64(),
            "long": pa.int64(),
            "short": pa.int16(),
            "byte": pa.int8(),
            "string": pa.string(),
            "float": pa.float32(),
            "double": pa.float64(),
            "boolean": pa.bool_(),
            "binary": pa.binary(),
            "date": pa.date32(),
            "timestamp": pa.timestamp("ns"),
            "timestamp_ntz": pa.timestamp("ns", tz=None),  # type: ignore
            "decimal": pa.decimal128(38, 10),
        }
        if schema in type_map:
            return type_map[schema]
        if schema.startswith("decimal"):
            parts = schema[schema.index("(") + 1 : schema.index(")")].split(",")
            precision = int(parts[0].strip())
            scale = int(parts[1].strip())
            return pa.decimal128(precision, scale)
        raise ValueError(f"unknown primitive type {schema}")
    if schema["type"] == "struct":
        for f in schema["fields"]:
            fields.append(
                pa.field(
                    f["name"],
                    to_pyarrow_schema(field_to_type(f)),
                    nullable=f.get("nullable", True),
                )
            )
        return pa.struct(fields)
    if schema["type"] == "array":
        return pa.list_(to_pyarrow_schema(schema["elementType"]))
    if schema["type"] == "map":
        return pa.map_(
            to_pyarrow_schema(schema["keyType"]),
            to_pyarrow_schema(schema["valueType"]),
        )
    raise ValueError(f"unknown complex type {schema}")


class Datasource:
    def __init__(
        self,
        version: str,
        tag: str,
        name: str,
        *,
        accounts: dict,
        config: DatasourceConfig,
        sql_context: ExecutionContext,
        basic_config: BasicConfig,
        df: Optional[ResultData] = None,
    ) -> None:
        self.version = version
        self.config = config
        self.tag = tag
        self.name = name
        self.df = df
        self.sql_context = sql_context
        self.basic_config = basic_config
        base_source_uri = SourceUri(
            config.uri,
            config.account,
            accounts,
            basic_config.data_path if config.file_type not in ["odbc"] else None,
            token_retrieval_func=basic_config.token_retrieval_func,
        )
        self.copy_local = config.copy_local
        self._execution_uri = None
        self.source_uri = base_source_uri

    def __str__(self) -> str:
        return str(self.source_uri)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        if self.df:
            self.df.__exit__(*args)

    @property
    def execution_uri(self):
        if not self.copy_local:
            return self.source_uri
        if self._execution_uri is None:
            self._execution_uri = self.source_uri.copy_to_local(
                os.path.join(
                    self.basic_config.local_data_cache_path,
                    hashlib.md5(self.source_uri.uri.encode("utf-8")).hexdigest(),
                ),
                self.config.file_type == "delta",
            )
        return self._execution_uri

    def file_exists(self):
        if self.config.file_type in ["odbc", "sqlite"]:
            return True  # the uri is not really a file here
        if not self.source_uri.exists():
            return False
        return True

    def get_delta_table(self, schema_only: bool):
        if self.config.file_type == "delta":
            if self.source_uri.exists():
                try:
                    df_uri, df_opts = (
                        self.source_uri if schema_only else self.execution_uri
                    ).get_uri_options(flavor="object_store")
                    return get_meta(PolarsEngine(df_opts), df_uri)
                except FileNotFoundError:
                    return None
        return None

    @property
    def tablename(self):
        if self.config.table_name:
            return self.config.table_name
        if self.version in ["1", "v1"]:
            return self.tag + "_" + self.name
        return self.tag + "_" + self.name + "_" + self.version

    @property
    def unique_table_name(self):
        if self.version in ["1", "v1"]:
            return self.tag + "_" + self.name
        return self.tag + "_" + self.name + "_" + self.version

    def get_table_name(self, force_unique_name: bool) -> ex.Table:
        tname = self.tablename if not force_unique_name else self.unique_table_name

        if "." in tname:
            parts = tname.split(".")
            if len(parts) == 3:
                return ex.Table(
                    this=ex.to_identifier(parts[2]),
                    db=ex.to_identifier(parts[1]),
                    catalog=ex.to_identifier(parts[0]),
                )
            assert len(parts) == 2
            return ex.Table(
                this=ex.to_identifier(parts[1]), db=ex.to_identifier(parts[0])
            )
        return ex.table_(tname)

    def get_schema(self) -> pa.Schema:
        schema: pa.Schema | None = None
        if self.config.file_type == "delta" and self.file_exists():
            dt = self.get_delta_table(schema_only=True)
            assert dt is not None and dt.schema is not None
            sc = to_pyarrow_schema(dt.schema)
            assert isinstance(sc, pa.StructType)
            return pa.schema(sc)
        if self.config.file_type == "parquet" and self.file_exists():
            fs, fs_uri = self.source_uri.get_fs_spec()
            schema = pyarrow.parquet.read_schema(fs_uri, filesystem=fs)
        if schema is not None:
            if self.config.select:
                fields = [
                    schema.field(item.name).with_name(item.alias or item.name)
                    for item in self.config.select
                ]
                return pyarrow.schema(fields)
            return schema
        return (self.get_df(endpoint="meta")).arrow_schema()

    def get_df(
        self,
        filters: Optional[FilterType] = None,
        endpoint: endpoints = "request",
    ) -> ResultData:
        if self.df is None:
            unique_table_name = (
                endpoint == "meta" and self.sql_context.supports_view_creation
            )
            query = select("*").from_(self.get_table_name(unique_table_name))

            self.query = (
                select_df(query, self.config.select, self.config.exclude or [])
                if endpoint != "query"
                else query
            )

            if self.df is None:
                self.sql_context.register_datasource(
                    self.unique_table_name if unique_table_name else self.tablename,
                    self.tablename,
                    self.source_uri if endpoint == "meta" else self.execution_uri,
                    self.config.file_type,
                    filters=filters,
                    meta_only=endpoint == "meta",
                )
                self.df = self.sql_context.execute_sql(self.query)

        return self.df  # type: ignore


def get_partition_filter(
    param: tuple[str, str],
    deltaMeta: DeltaMetadata,
    param_def: Iterable[Param | str],
) -> Optional[Tuple[str, OperatorType, Any]]:
    operators = ("<=", ">=", "=", "==", "in", "not in")
    key, value = param
    if not key or not value or key in ("limit", "offset"):
        return None

    prmdef_and_op = get_param_def(key, param_def)
    if prmdef_and_op is None:
        raise ValueError(f"thats not parameter: {key}")
    prmdef, op = prmdef_and_op
    if op not in get_args(DeltaOperatorTypes):
        return None
    colname = prmdef.colname or prmdef.name

    value_for_partitioning = value
    col_for_partitioning: Optional[str] = None
    deltaSchema = deltaMeta.schema
    assert deltaMeta.last_metadata is not None
    assert deltaSchema is not None
    for partcol in deltaMeta.last_metadata.get("partitionColumns", []):
        if partcol == colname:
            d_type = next(
                (f["type"] for f in deltaSchema["fields"] if f["name"] == partcol)
            )

            col_for_partitioning = partcol
            value_for_partitioning = (
                int(value) if d_type in ["int", "long", "integer"] else str(value)
            )

        elif partcol.startswith(colname + "_md5_prefix_"):
            col_for_partitioning = partcol
            prefix_len = int(partcol[len(colname + "_md5_prefix_") :])
            if isinstance(value, (List, Tuple)):
                hashvl = [hashlib.md5(str(v).encode("utf8")).hexdigest() for v in value]
                value_for_partitioning = [hvl[:prefix_len] for hvl in hashvl]
            else:
                hashvl = hashlib.md5(str(value).encode("utf8")).hexdigest()
                value_for_partitioning = hashvl[:prefix_len]
            if op not in operators:
                col_for_partitioning = None
                continue
        elif partcol.startswith(colname + "_md5_mod_"):
            col_for_partitioning = partcol
            modulo_len = int(partcol[len(colname + "_md5_mod_") :])
            if isinstance(value, (List, Tuple)):
                hashvl = [
                    int(hashlib.md5(str(v).encode("utf8")).hexdigest(), 16)
                    for v in value
                ]
                value_for_partitioning = [str(hvl % modulo_len) for hvl in hashvl]
            else:
                hashvl = int(hashlib.md5(str(value).encode("utf8")).hexdigest(), 16)
                value_for_partitioning = str(hashvl % modulo_len)
            if op not in operators:
                col_for_partitioning = None
                continue
        elif partcol.startswith(colname + "_prefix_"):
            col_for_partitioning = partcol
            prefix_len = int(partcol[len(colname + "_prefix_") :])
            if isinstance(value, (List, Tuple)):
                value_for_partitioning = [v[:prefix_len] for v in value]
            else:
                value_for_partitioning = value[:prefix_len]
            if op not in operators:
                col_for_partitioning = None
                continue

    if not col_for_partitioning:
        return None
    # partition value must be string
    return (
        col_for_partitioning,
        op,
        value_for_partitioning,
    )


def filter_partitions_based_on_params(
    deltaMeta: DeltaMetadata,
    params: dict,
    param_def: Iterable[Param | str],
):
    if not deltaMeta.last_metadata:
        return None
    if not deltaMeta.last_metadata.get("partitionColumns"):
        return None

    partition_filters = []
    results = [
        get_partition_filter(param, deltaMeta, param_def) for param in params.items()
    ]
    partition_filters = [result for result in results if result is not None]

    return partition_filters if len(partition_filters) > 0 else None


ExpType: TypeAlias = "Sequence[Union[ex.Binary, ex.Condition]]"


def concat_expr(
    exprs: ExpType,
) -> "Union[ex.Binary,  ex.Condition]":
    expr: Optional[ex.Binary | ex.Condition] = None
    for e in exprs:
        if expr is None:
            expr = e
        else:
            expr = expr.__and__(e)
    assert expr is not None
    return expr


def _create_inner_expr(
    columns: Optional[List[str]],
    prmdef,
    e,
):
    inner_expr: Optional[ex.Binary] = None
    for ck, cv in e.items():
        logger.debug(f"key = {ck}, value = {cv}, columns = {columns}")
        if (columns and ck not in columns) and ck not in prmdef.combi:
            pass
        else:
            if inner_expr is None:
                inner_expr = (
                    ex.column(ck, quoted=True).eq(cv)
                    if (cv or cv == 0)
                    else ex.column(ck, quoted=True).is_(ex.Null())
                )
            else:
                inner_expr = inner_expr & (
                    ex.column(ck, quoted=True).eq(cv)
                    if cv or cv == 0
                    else ex.column(ck, quoted=True).is_(ex.Null())
                )
    return inner_expr


def _sql_value(value: str | datetime | date | None, engine: str):
    if engine != "polars":
        return value  # all other engines convert automatically

    if isinstance(value, datetime):
        return ex.cast(ex.convert(value), "datetime")
    if isinstance(value, date):
        return ex.cast(ex.convert(value), "date")
    return value


def get_filters(
    params: dict[str, Any],
    param_def: list[Union[Param, str]],
    columns: Optional[Sequence[str]],
) -> tuple[FilterType, bool]:
    result: list[tuple[str, SupportedOperator, Any]] = []
    covered_all = True
    for key, value in params.items():
        if not key or not value or key in ("limit", "offset"):
            continue  # can that happen? I don't know
        prmdef_and_op = get_param_def(key, param_def)
        if prmdef_and_op is None:
            continue
        prmdef, op = prmdef_and_op
        colname = prmdef.colname or prmdef.name

        if prmdef.combi:
            covered_all = False
            continue
        elif columns and colname not in columns:
            covered_all = False
            pass
        else:
            if op not in get_args(SupportedOperator):
                covered_all = False
            else:
                result.append((colname, cast(SupportedOperator, op), value))

    return result, covered_all


def filter_df_based_on_params(
    context: ExecutionContext,
    params: dict[str, Any],
    param_def: list[Union[Param, str]],
    columns: Optional[list[str]],
) -> Optional[ex.Condition | ex.Binary]:
    expr: Optional[ex.Condition] = None
    exprs: list[ex.Condition] = []

    for key, value in params.items():
        if not key or not value or key in ("limit", "offset"):
            continue  # can that happen? I don't know
        prmdef_and_op = get_param_def(key, param_def)
        if prmdef_and_op is None:
            continue
        prmdef, op = prmdef_and_op
        colname = prmdef.colname or prmdef.name

        if prmdef.combi:
            outer_expr: Optional[ex.Condition] = None
            results = [_create_inner_expr(columns, prmdef, e) for e in value]
            for inner_expr in results:
                if inner_expr is not None:
                    if outer_expr is None:
                        outer_expr = inner_expr
                    else:
                        outer_expr = outer_expr | (inner_expr)
            if outer_expr is not None:
                exprs.append(outer_expr)

        elif columns and colname not in columns:
            pass

        else:
            match op:
                case "<":
                    exprs.append(
                        ex.column(colname, quoted=True)
                        < _sql_value(value, engine=context.engine_name)
                    )
                case ">":
                    exprs.append(
                        ex.column(colname, quoted=True)
                        > _sql_value(value, engine=context.engine_name)
                    )
                case ">=":
                    exprs.append(
                        ex.column(colname, quoted=True)
                        >= _sql_value(value, engine=context.engine_name)
                    )
                case "<=":
                    exprs.append(
                        ex.column(colname, quoted=True)
                        <= _sql_value(value, engine=context.engine_name)
                    )
                case "<>":
                    exprs.append(
                        ex.column(colname, quoted=True).neq(
                            _sql_value(value, engine=context.engine_name)
                        )
                        if value is not None
                        else ~ex.column(colname, quoted=True).is_(ex.convert(None))
                    )
                case "=":
                    exprs.append(
                        ex.column(colname, quoted=True).eq(
                            _sql_value(value, engine=context.engine_name)
                        )
                        if value is not None
                        else ex.column(colname, quoted=True).is_(ex.convert(None))
                    )
                case "not contains":
                    exprs.append(
                        context.term_like(
                            ex.column(colname, quoted=True), value, "both", negate=True
                        )
                    )
                case "contains":
                    exprs.append(
                        context.term_like(
                            ex.column(colname, quoted=True), value, "both"
                        )
                    )
                case "startswith":
                    exprs.append(
                        context.term_like(ex.column(colname, quoted=True), value, "end")
                    )
                case "has":
                    exprs.append(
                        ex.func(
                            context.array_contains_func,
                            ex.column(colname, quoted=True),
                            ex.convert(value),
                        )
                    )
                case "in":
                    lsv = cast(list[str], value)
                    if len(lsv) > 0:
                        exprs.append(ex.column(colname, quoted=True).isin(*lsv))
                case "not in":
                    lsv = cast(list[str], value)
                    if len(lsv) > 0:
                        exprs.append(~ex.column(colname, quoted=True).isin(*lsv))
                case "between":
                    lsv = cast(list[str], value)
                    if len(lsv) == 2:
                        exprs.append(
                            ex.column(colname, quoted=True).between(lsv[0], lsv[1])
                        )
                    else:
                        from fastapi import HTTPException

                        raise HTTPException(
                            400, "Must have an array with 2 elements for between"
                        )
                case "not between":
                    lsv = cast(list[str], value)
                    if len(lsv) == 2:
                        exprs.append(
                            ~ex.column(colname, quoted=True).between(lsv[0], lsv[1])
                        )
                    else:
                        from fastapi import HTTPException

                        raise HTTPException(
                            400, "Must have an array with 2 elements for between"
                        )

                case operator:
                    logger.error(f"wrong parameter for filter {operator}")
    if len(exprs) == 0:
        return None
    return concat_expr(exprs)



from dataclasses import dataclass
from fastapi import FastAPI
from bmsdna.lakeapi.core.config import BasicConfig, Configs, get_default_config
from bmsdna.lakeapi.core.route import init_routes
import os


@dataclass(frozen=True)
class LakeApiStartInfo:
    start_config: BasicConfig
    config: Configs


def init_lakeapi(
    app: FastAPI,
    use_basic_auth: bool,
    start_config: BasicConfig | None = None,
    config: Configs | str | None = None,
) -> LakeApiStartInfo:
    start_config = start_config or get_default_config()
    real_config: Configs
    if config is None:
        real_config = Configs.from_yamls(
            start_config, os.getenv("CONFIG_PATH", "config.yml")
        )
    elif isinstance(config, str):
        real_config = Configs.from_yamls(start_config, config)
    else:
        real_config = config
    router = init_routes(real_config, start_config)
    if use_basic_auth:
        from bmsdna.lakeapi.core.uservalidation import add_user_middlware

        add_user_middlware(app, start_config, real_config.users)

    app.include_router(router)
    return LakeApiStartInfo(start_config, real_config)



import time
import sys
import asyncio
from fastapi.concurrency import run_in_threadpool
from fastapi.testclient import TestClient
import pytest

sys.path.append(".")


engines = ("duckdb", "polars")


@pytest.mark.asyncio
async def test_async_execution(client: TestClient):
    # simulate a lot of async requests and guarantee that the execution is under 1 second

    def call_api_1(engine, format):
        start = time.time()
        response = client.get(
            f"/api/v1/test/fake_delta?limit=10000&format=json&%24engine={engine}&format={format}"
        )
        end = time.time()
        assert response.status_code == 200

        duration = end - start
        print(f"Engine {engine} took {duration} seconds with format {format}")
        assert duration < 1.0

    def call_api_2(engine, format):
        start = time.time()
        response = client.get(
            f"/api/v1/test/fruits_partition?limit=10000&format=json&cars=audi&%24engine={engine}&format={format}"
        )
        end = time.time()
        assert response.status_code == 200

        duration = end - start
        print(f"Engine {engine} took {duration} seconds with format {format}")
        assert duration < 1.0

    def call_api_3(engine, format):
        start = time.time()
        response = client.get(
            f"/api/v1/test/fruits_sortby_desc?limit=10000&format=json&cars=audi&%24engine={engine}&format={format}"
        )
        end = time.time()
        assert response.status_code == 200

        duration = end - start
        print(f"Engine {engine} took {duration} seconds with format {format}")
        assert duration < 1.0

    def call_api_4(engine, format):
        start = time.time()
        response = client.get(
            f"/api/v1/test/fruits_sortby_asc?limit=10000&format=json&cars=audi&%24engine={engine}&format={format}"
        )
        end = time.time()
        assert response.status_code == 200

        duration = end - start
        print(f"Engine {engine} took {duration} seconds with format {format}")
        assert duration < 1.0

    tasks = []
    for _ in range(100):
        for e in engines:
            for f in ("json", "csv", "arrow", "parquet"):
                tasks.append(run_in_threadpool(call_api_1, e, f))
                tasks.append(run_in_threadpool(call_api_2, e, f))
                tasks.append(run_in_threadpool(call_api_3, e, f))
                tasks.append(run_in_threadpool(call_api_4, e, f))
    print(f"Running {len(tasks)} tasks")
    await asyncio.gather(*tasks)
